#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import io
import operator
import os
import shlex
import stat
import subprocess
import threading
from repeatfs.descriptor_entry import DescriptorEntry


class ProcessIO():
    """ Provides an interface for process IO with a blocking buffer """
    def __init__(self, cache_entry):
        self.cache_entry = cache_entry
        self.process = None
        self.pid_auth = dict()
        self.stream_buffer = None
        self.write_open = True
        self.blocks_byte_pos = 0
        self.reset_pos = 0
        self.lock = threading.Condition()
        self.read_active = False
        self.write_active = False

    def _read_buffer(self, size):
        # Must be called with lock
        block_size = self.cache_entry.core.configuration.values["block_size"]

        try:
            # Directly pass-through BufferedReader (stdout/stderr) streams
            if isinstance(self.stream_buffer, io.BufferedReader):
                ret_data = self.stream_buffer.read(size)
                return ret_data

            # Block until buffer filled or all write modes from source process are (currently) closed
            while True:
                current_size = len(self.stream_buffer.getbuffer())
                if current_size >= block_size or not self.write_open: break
                self.lock.wait()

            # Reset to end of last read
            self.stream_buffer.seek(self.reset_pos)
            ret_data = self.stream_buffer.read(size)

            # If we've read entire block, reset
            if self.stream_buffer.tell() == block_size:
                self.stream_buffer.seek(0)
                self.stream_buffer.truncate()

            # Mark current position as next reset
            self.reset_pos = self.stream_buffer.tell()

            return ret_data

        finally:
            self.lock.notifyAll()

    def _write_buffer(self, data):
        # Must be called with lock
        block_size = self.cache_entry.core.configuration.values["block_size"]

        try:
            # Writes are only valid for BytesIO (file) streams
            if not isinstance(self.stream_buffer, io.BytesIO): return

            # Block until buffer is not full
            while True:
                current_size = len(self.stream_buffer.getbuffer())
                if current_size < block_size: break
                self.lock.wait()

            return self.stream_buffer.write(data)

        finally:
            self.lock.notifyAll()

    # Prepare an internally generated file
    def _internal_prepare(self):
        sys_config = self.cache_entry.core.configuration.values
        file_config = self.cache_entry.core.configuration.actions[self.cache_entry.file_entry.virt_action[:2]]

        if file_config["internal"] is None: return

        # Create the temporary file for the entry output
        path = os.path.join(sys_config["cache_path"], "{0}.temp".format(self.cache_entry.cache_path))
        with os.fdopen(os.open(path, os.O_CREAT | os.O_WRONLY), "ab") as handle:
            method = file_config["internal"][0].split(".")
            operator.attrgetter(".".join(method[1:]))(self)(self, handle, *file_config["internal"][1:])

    # Cleanup internally generated files
    def _internal_cleanup(self):
        sys_config = self.cache_entry.core.configuration.values
        path = os.path.join(sys_config["cache_path"], "{0}.temp".format(self.cache_entry.cache_path))

        if os.path.exists(path):
            os.remove(path)

    # Run command associated with file
    def req_init(self):
        # Must be called with lock
        sys_config = self.cache_entry.core.configuration.values
        file_config = self.cache_entry.core.configuration.actions[self.cache_entry.file_entry.virt_action[:2]]
        match_groups = self.cache_entry.file_entry.virt_action[2]

        # Ignore init request if running or previously completed
        if self.cache_entry.final or self.process: return

        # Reset stream position
        self.blocks_byte_pos = 0

        # Execute internal prepration, if applicable
        self._internal_prepare()

        if self.cache_entry.file_entry.file_type == stat.S_IFREG:
            # Build string replacements
            replacements = dict()
            replacements["input"] = self.cache_entry.file_entry.derived_source.paths["abs_mount"]
            replacements["output"] = self.cache_entry.file_entry.paths["abs_mount"]
            replacements["output_base"] = replacements["output"][:-len(file_config["ext"])]
            replacements["temp"] = "{0}.temp".format(self.cache_entry.cache_path)

            for group_pair in enumerate(match_groups):
                group_idx = "input_{0}".format(group_pair[0])
                replacements[group_idx] = os.path.join(os.path.dirname(replacements["input"]), group_pair[1])

            # Construct the command
            output = file_config["output"]
            command = file_config["cmd"]
            command = command.format(**replacements)
            self.cache_entry.core.log("Running command \"{0}\", directing to {1}".format(command, output), self.cache_entry.core.LOG_DEBUG)

            # Start execution
            stdout_dev = subprocess.PIPE if output == "stdout" else subprocess.DEVNULL
            stderr_dev = subprocess.PIPE if output == "stderr" else subprocess.DEVNULL
            self.process = subprocess.Popen(shlex.split(command), bufsize=sys_config["block_size"], stdout=stdout_dev, stderr=stderr_dev)
            self.pid_auth[self.process.pid] = True

            # Standard streams are pass-through to a file ReadBuffer
            if output == "stdout": self.stream_buffer = self.process.stdout
            if output == "stderr": self.stream_buffer = self.process.stderr
            if output == "file": self.stream_buffer = io.BytesIO(bytes(0))

    # End process if running
    def end_process(self):
        # Must be called with lock
        if self.process is not None:
            self.cache_entry.core.log("Killing process {0} (may already be killed)".format(self.process.pid), self.cache_entry.core.LOG_DEBUG)
            self.process.kill()
            self.process.wait()
            self.write_open = False
            self.process = None
            self.pid_auth.clear()

            # Cleanup internal preparation, if applicable
            self._internal_cleanup()

    # Check if the process is complete
    def check_process(self):
        # Must be called with lock
        if self.process:
            if self.process.poll() is not None:
                self.cache_entry.core.log("Process finalized: {0} {1} {2}".format(self.cache_entry.file_entry.paths["abs_virt"], self.cache_entry.mtime, self.cache_entry.file_entry.virt_mtime), self.cache_entry.core.LOG_DEBUG)
                # self.size = (self.process_next_block - 1) * self.core.configuration.values["block_size"] + block_size
                self.cache_entry.mtime = self.cache_entry.file_entry.virt_mtime
                self.cache_entry.final = True

    def check_lineage(self, pid):
        """ Check to see if pid is in process owner pid lineage """
        self.pid_auth[pid] = False
        current_pid = pid

        while current_pid > 1:
            # Check if current PID in lineage is the owner
            if current_pid == self.process.pid:
                self.pid_auth[pid] = True
                break

            with open("/proc/{0}/stat".format(current_pid), "r") as handle:
                stat_info = handle.readline().split(" ")

                # Find end of process name
                for field_mod in range(len(stat_info) - 1):
                    if stat_info[1 + field_mod].endswith(")"): break

                current_pid = int(stat_info[3 + field_mod])

    # Check if descriptor's open FUSE context (at open) is process owner
    def context_owner(self, descriptor=None, pid=None):
        if descriptor is not None:
            desc_entry = DescriptorEntry.get(descriptor)
            context_pid = desc_entry.open_pid
        elif pid is not None:
            context_pid = pid
        else:
            return False

        if not self.process: return False

        # If context pid isn't recorded in authorization rules, check and update
        if context_pid not in self.pid_auth:
            self.check_lineage(context_pid)

        return self.pid_auth[context_pid]

    # Perform a stream read if available
    def read(self, req_block):
        sys_config = self.cache_entry.core.configuration.values
        block_size = sys_config["block_size"]

        with self.lock:
            while self.read_active:
                self.lock.wait()

            try:
                # Ensure no other reads start when lock is released during blocked read
                self.read_active = True

                process_block = self.blocks_byte_pos // block_size
                process_start = self.blocks_byte_pos % block_size
                process_data = None

                # If process is still running and requested block is at or after process position, read stream up to end of block
                if self.process and req_block >= process_block:
                    process_data = bytearray(self._read_buffer(block_size - process_start))
                    self.blocks_byte_pos += len(process_data)

                    # If that was the final byte in stream, check if process complete
                    if len(process_data) == 0:
                        self.check_process()

                return (process_block, process_start, process_data)

            finally:
                self.read_active = False
                self.lock.notifyAll()

    # Perform a stream write if available, and return length not sent to stream
    def write(self, data, pos, descriptor):
        sys_config = self.cache_entry.core.configuration.values
        block_size = sys_config["block_size"]

        with self.lock:
            while self.write_active:
                self.lock.wait()

            try:
                # Ensure no other writes start when lock is released during blocked read
                self.write_active = True

                # Do not write to stream if not our process or process is complete
                if not self.context_owner(descriptor=descriptor): return len(data)

                # Note portion not sent to stream
                ret_len = self.blocks_byte_pos - pos
                if ret_len > len(data):
                    ret_len = len(data)

                # Send remainder to stream
                if ret_len < len(data):
                    # Fill zeroes if position is after current tell position of buffer
                    abs_tell_pos = self.stream_buffer.tell() + ((self.blocks_byte_pos // block_size) * block_size)
                    if pos > abs_tell_pos:
                        self.truncate(pos, descriptor, True)

                    # Adjust buffer write amount for portion returned for mem cache write
                    buffer_remain = len(data)
                    if pos < self.blocks_byte_pos:
                        buffer_remain -= (self.blocks_byte_pos - pos)
                        self.stream_buffer.seek(0)
                    else:
                        # Position must be within blocks_byte_pos and tell at this point
                        blocks_byte_start = ((self.blocks_byte_pos // block_size) * block_size)
                        self.stream_buffer.seek(pos - blocks_byte_start)

                    while buffer_remain > 0:
                        write_len = buffer_remain
                        if write_len > (block_size - self.stream_buffer.tell()):
                            write_len = (block_size - self.stream_buffer.tell())

                        data_pos = len(data) - buffer_remain
                        self._write_buffer(data[data_pos:data_pos + write_len])
                        buffer_remain -= write_len

                return ret_len

            finally:
                self.write_active = False
                self.lock.notifyAll()

    # Perform a stream truncate if available, return False if truncate should happen in memory cache
    def truncate(self, pos, descriptor, write_call):
        sys_config = self.cache_entry.core.configuration.values
        block_size = sys_config["block_size"]

        with self.lock:
            while self.write_active and not write_call:
                self.lock.wait()

            try:
                # Do not write to stream if not our process or process is complete
                if not self.context_owner(descriptor=descriptor): return False

                # Check if truncate should happen in memory cache
                if pos < self.blocks_byte_pos: return False

                # Calculate truncate position relative to buffer
                trunc_remain = pos - ((self.blocks_byte_pos // block_size) * block_size)

                if trunc_remain < self.stream_buffer.tell():
                    # If before current buffer position, truncate
                    self.stream_buffer.truncate(trunc_remain)
                else:
                    # Otherwise, zero fill
                    trunc_remain -= self.stream_buffer.tell()

                    while trunc_remain > 0:
                        empty_len = trunc_remain
                        if empty_len > (block_size - self.stream_buffer.tell()):
                            empty_len = (block_size - self.stream_buffer.tell())

                        self._write_buffer(bytes(empty_len))
                        trunc_remain -= empty_len

                return True

            finally:
                if not write_call:
                    self.write_active = False
                    self.lock.notifyAll()

    # Close the stream
    def close(self, read, write):
        with self.lock:
            # If no more open reads, end the process
            if read:
                self.end_process()

            # If no more open owner writes, close the stream
            if write:
                self.write_open = False
                self.lock.notifyAll()



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import hashlib
import io
import json
import operator
import os
import queue
import subprocess
import threading
import time
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.process_io import ProcessIO


class CacheEntry:
    """ Provides functionality for caching virtual files to memory and disk """
    BLOCK_DATA, BLOCK_DIRTY = range(2)
    IO_READ, IO_WRITE, IO_TRUNCATE, IO_RESET = range(4)

    entries = dict()
    block_history = queue.Queue()

    @classmethod
    def api_config(cls, api_out):
        """ API: Receive configure command """
        # Attempt to configure VDF
        try:
            # Check and clean path
            for field in ("path", "options"):
                if field not in api_out.cmd_info:
                    api_out.respond(status="error", message="{} not specified".format(field))
                    return

            path = api_out.cmd_info["path"].replace("|", "/")
            if path not in CacheEntry.entries:
                    api_out.respond(status="error", message="invalid path specified")
                    return

            # Check, clean, and update configuration
            options = api_out.cmd_info["options"]
            if "expand_procs" in options:
                options["expand_procs"] = [tuple(val.split("|")) for val in options["expand_procs"]]

            cache_entry = CacheEntry.entries[path]
            cache_entry.update_config(api_out.cmd_info["options"])

            # Clear cache
            cache_entry.io(CacheEntry.IO_RESET, 0, None, 1, api_out.desc_id)

            api_out.respond(status="ok", message="", final=True)

        except json.decoder.JSONDecodeError:
            api_out.respond(status="malformed", message=api_out.cmd_info["options"])

        except Exception as e:
            api_out.respond(status="error", message=e)

    def __init__(self, core, entry):
        self.core = core

        # One-time intialization
        self.lock = threading.Condition()
        self.blocks = dict()
        self.waiting = dict()
        self.config = dict()
        self.descriptors = set()
        self.process_io = ProcessIO(self)

        # Reset-time intialization: blocks, blocks_block_pos, file_entry, size, mtime, final, cache_path
        self.reset_cache(entry)

    def _io_read(self, block_idx, block_pos, size, ret_data, ret_size):
        # Must be called with lock
        block_data = self.blocks[block_idx][self.BLOCK_DATA]

        # Calculate the amount that can be retrieved from this block
        consume_size = size - ret_size
        avail_size = len(block_data[block_pos:block_pos + consume_size])
        if consume_size > avail_size: consume_size = avail_size

        # Retrieve data
        ret_data[ret_size:ret_size + consume_size] = self.blocks[block_idx][self.BLOCK_DATA][block_pos:block_pos + consume_size]
        ret_size += consume_size

        return ret_data, ret_size

    def _io_write(self, block_idx, block_pos, new_data, new_dirty, ret_size):
        # Must be called with lock
        block_size = self.core.configuration.values["block_size"]

        block_data = bytearray()
        if block_idx in self.blocks:
            block_data = self.blocks[block_idx][self.BLOCK_DATA]

        # Calculate the amount that can be written to this block
        consume_size = len(new_data) - ret_size
        avail_size = block_size - block_pos
        if consume_size > avail_size: consume_size = avail_size

        # Write data
        block_data[block_pos:block_pos + consume_size] = new_data[ret_size:ret_size + consume_size]
        self.blocks[block_idx] = (block_data, new_dirty)
        ret_size += consume_size

        # Note write in history
        self.block_history.put((self, block_idx))

        # Update current block and file size
        if block_idx + 1 > self.blocks_block_pos:
            self.blocks_block_pos = block_idx + 1

        # Update file size if this is the last block
        if block_idx + 1 == self.blocks_block_pos:
            self.size = block_idx * block_size + len(block_data)

        return ret_size

    def _io_truncate(self, block_idx, block_pos, ret_size):
        # Must be called with lock
        block_size = self.core.configuration.values["block_size"]

        # Delete from position onward
        for del_idx in range(block_idx + 1, self.blocks_block_pos):
            del self.blocks[del_idx]

        # Delete up to position in last block, mark dirty
        self.blocks[block_idx][self.BLOCK_DATA][:] = self.blocks[block_idx][self.BLOCK_DATA][:block_pos]
        self.blocks[block_idx] = (self.blocks[block_idx][self.BLOCK_DATA], True)

        # Update block position and size
        self.blocks_block_pos = block_idx + 1
        self.size = block_idx * block_size + block_pos

        return ret_size

    # TODO: Make 0 blocks "virtual", note only - do not create bytearrays/take up memory
    def _io_fill(self, block_idx, block_pos):
        # Must be called with lock
        block_size = self.core.configuration.values["block_size"]

        # Fill blocks between current and end
        for fill_idx in range(self.blocks_block_pos - 1, block_idx):
            # No previous blocks if file is empty
            if fill_idx == -1: continue

            if fill_idx == self.blocks_block_pos - 1:
                # Fill the remainder of current last block
                block_data = self.blocks[fill_idx][self.BLOCK_DATA]
                block_data.extend(bytearray(block_size - len(block_data)))
                self.blocks[fill_idx] = (block_data, True)
            else:
                # Create new block
                self.blocks[fill_idx] = (bytearray(block_size), True)

        # Fill in bytes in new last block
        if block_idx not in self.blocks:
            block_data = bytearray()
        else:
            block_data = self.blocks[block_idx][self.BLOCK_DATA]

        block_data.extend(bytearray(block_pos - len(block_data)))
        self.blocks[block_idx] = (block_data, True)

        # Update block position and size
        self.blocks_block_pos = block_idx + 1
        self.size = block_idx * block_size + block_pos

    def io(self, operation, pos, data, size, descriptor):
        sys_config = self.core.configuration.values
        desc_entry = DescriptorEntry.get(descriptor)
        ret_data = bytearray(size)
        ret_size = 0

        while ret_size < size:
            block_size = sys_config["block_size"]
            block = (pos + ret_size) // block_size
            start = (pos + ret_size) % block_size

            # Phase 1 (Availability, Resets) - wait for IO priority
            self.core.log("IO loop phase 1: op {0}, block {1} start {2} ret_size {3}".format(operation, block, start, ret_size), self.core.LOG_DEBUG)
            with self.lock:
                self.priority_wait(block, descriptor, operation)

                try:
                    # Immediately process cache resets, then exit
                    if operation == self.IO_RESET:
                        self.reset_cache(desc_entry.file_entry)
                        return size

                    # Request a block if not available or not full (may also finalize file)
                    req_block = (block not in self.blocks) or len(self.blocks[block][self.BLOCK_DATA]) < block_size
                finally:
                    self.lock.notify_all()

            # Phase 2 (Fetch from stream/disk into memory cache)
            if req_block:
                self.core.log("IO loop phase 2", self.core.LOG_DEBUG)
                self.check_expired()
                self.req_mem_block(block, descriptor, operation)

            # Phase 3 (Perform partial and full IO to/from memory cache)
            self.core.log("IO loop phase 3", self.core.LOG_DEBUG)
            with self.lock:
                self.priority_wait(block, descriptor, operation)

                try:
                    # Handle partial blocks
                    if operation == self.IO_READ:
                        # For reads, return available data if past EOF
                        if self.final and (pos + ret_size) >= self.size:
                            return bytes(ret_data[:ret_size])

                    if operation == self.IO_WRITE or operation == self.IO_TRUNCATE:
                        # For writes, grow file past EOF
                        if pos > self.size:
                            self._io_fill(block, start)

                    # If block is now available, perform IO on it
                    if block in self.blocks:
                        # Perform IO operation
                        if operation == self.IO_READ:
                            ret_data, ret_size = self._io_read(block, start, size, ret_data, ret_size)

                        if operation == self.IO_WRITE:
                            ret_size = self._io_write(block, start, data, True, ret_size)

                        if operation == self.IO_TRUNCATE:
                            ret_size = self._io_truncate(block, start, size)

                finally:
                    self.core.log("IO loop complete, notifying all", self.core.LOG_DEBUG)
                    self.lock.notify_all()

        return bytes(ret_data) if operation == self.IO_READ else ret_size

    # Is descriptor in read mode
    def is_descriptor_read(self, descriptor):
        desc_entry = DescriptorEntry.get(descriptor)
        return ((desc_entry.flags % 2) == 0)

    # Is descriptor in write mode
    def is_descriptor_write(self, descriptor):
        desc_entry = DescriptorEntry.get(descriptor)
        return ((desc_entry.flags & 0x3) > 0)

    # Register a descriptor, check for owner writes
    def register_descriptor(self, descriptor):
        desc_entry = DescriptorEntry.get(descriptor)
        with self.lock:
            # Register descriptor in set
            self.core.log("Registering descriptor {0}".format(descriptor), self.core.LOG_DEBUG)
            self.descriptors.add(descriptor)

            # Check for owner write
            desc_write = self.process_io.context_owner(descriptor=descriptor)
            desc_write &= self.core.is_flag_write(desc_entry.flags)

            if desc_write:
                self.process_io.write_open = True

    # Unregister descriptor, check for last read/write
    def unregister_descriptor(self, descriptor):
        reads = False
        writes = False

        with self.lock:
            # Unregister descriptor from set and waiting
            self.descriptors.remove(descriptor)
            self.waiting.pop(descriptor, None)

            # For for remaining reads/writes
            for desc_remain in self.descriptors:
                desc_entry = DescriptorEntry.get(desc_remain)

                desc_read = self.core.is_flag_read(desc_entry.flags)
                reads |= desc_read

                desc_write = self.process_io.context_owner(descriptor=desc_entry.id)
                desc_write &= self.core.is_flag_write(desc_entry.flags)
                writes |= desc_write

            # Close stream if no further reads or writes
            self.process_io.close(not reads, not writes)
            self.lock.notifyAll()

    # Reset file's cache
    def reset_cache(self, file_entry):
        # Must be called with lock (or from constructor)

        # Initialize memory cache
        self.blocks.clear()
        self.blocks_block_pos = 0
        self.file_entry = file_entry
        self.size = 0
        self.mtime = 0
        self.final = False

        # Compute disk cache location
        hash = hashlib.md5()
        hash.update(self.file_entry.paths["abs_real"].encode("utf8"))
        self.cache_path = os.path.join(self.core.configuration.values["cache_path"], hash.hexdigest())

        # Initialize disk cache
        cache_handle = open(self.cache_path, "w")
        cache_handle.close()

    # Flush a block to disk cache
    def flush_block(self, block_idx):
        # Must be called with lock
        block_size = self.core.configuration.values["block_size"]
        block_data = self.blocks[block_idx][self.BLOCK_DATA]
        block_dirty = self.blocks[block_idx][self.BLOCK_DIRTY]

        # Skip flush if disk_cache is disabld or clean
        if not block_dirty: return

        end = (block_idx * block_size + len(block_data)) == self.size
        self.set_disk_block(block_idx, block_data, end)

    # Flush expired blocks from memory to disk cache
    def check_expired(self):
        block_size = self.core.configuration.values["block_size"]
        max_blocks = self.core.configuration.values["store_size"] // block_size

        if self.block_history.qsize() >= max_blocks:
            self.core.log("Initiating cache flush", self.core.LOG_DEBUG)
            while self.block_history.qsize() > max_blocks // 2:
                entry_pair = self.block_history.get()

                with entry_pair[0].lock:
                    if entry_pair[1] in entry_pair[0].blocks:
                        # Flush dirty blocks to disk cache
                        entry_pair[0].flush_block(entry_pair[1])

                        # Delete block from memory cache
                        del entry_pair[0].blocks[entry_pair[1]]

    # Wait until block operation has priority
    def priority_wait(self, block, descriptor, operation):
        sys_config = self.core.configuration.values

        # Must be called with lock
        while True:
            # Only prioritize reads (for now)
            if operation != self.IO_READ: return

            self.waiting[descriptor] = (block, time.time())

            # Priority to reads requesting blocks already cached
            if block in self.blocks: return

            # Priority to reads early in file (without timeout)
            min_block = block
            for block_pair in self.waiting.values():
                if time.time() - block_pair[1] < sys_config["read_timeout"]:
                    if block_pair[0] < min_block:
                        min_block = block_pair[0]

            if block == min_block: return

            self.lock.wait(sys_config["read_timeout"])

    # Get disk block
    def get_disk_block(self, block_idx):
        block_size = self.core.configuration.values["block_size"]
        file_size = os.path.getsize(self.cache_path)

        with open(self.cache_path, "rb") as cache_handle:
            cache_handle.seek(block_size * block_idx)
            block_data = cache_handle.read(block_size)
            end = (cache_handle.tell() == file_size)

            return (block_data, end)

    # Set disk block
    def set_disk_block(self, block_idx, block_data, end):
        byte_pos = self.core.configuration.values["block_size"] * block_idx
        file_size = os.path.getsize(self.cache_path)

        with open(self.cache_path, "r+b") as cache_handle:
            # Ensure data exists up to location
            if file_size < byte_pos:
                cache_handle.seek(0, 2)
                cache_handle.write(bytearray(byte_pos - file_size))

            # Write block
            cache_handle.seek(byte_pos, 0)
            cache_handle.write(block_data)

            # Truncate if last block
            if end: cache_handle.truncate()

    # Load a block into the memory cache using requested block as hint
    def req_mem_block(self, req_block, descriptor, operation):
        # Must be called with lock
        block_size = self.core.configuration.values["block_size"]

        # Request process init (will ignore if already running/complete)
        self.process_io.req_init()

        # Attempt to load from disk cache before reading from the process
        disk_data = None

        with self.lock:
            # If requested block is before current block (and wasn't already fetched), fetch from disk
            if req_block not in self.blocks and req_block < self.blocks_block_pos:
                disk_data = self.get_disk_block(req_block)[0]
                self._io_write(req_block, 0, disk_data, False, 0)

                # If this wasn't a full block, also attempt process IO
                if len(disk_data) == block_size:
                    return

        # For read operations from non-owner, if not in the disk cache, fetch more from process
        if not self.process_io.context_owner(descriptor=descriptor) and operation == self.IO_READ:
            process_info = self.process_io.read(req_block)
            process_block, process_start, process_data = process_info

            if process_data:
                with self.lock:
                    self._io_write(process_block, process_start, process_data, True, 0)

    def update_config(self, options):
        """ Update cache entry configuration """
        with self.lock:
            self.config.update(options)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import time
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry
from repeatfs.provenance.file_record import FileRecord
from repeatfs.provenance.process_record import ProcessRecord


class IORecord:
    """ Provides IO information """
    IO_READ, IO_WRITE = range(2)
    IO_START, IO_END = range(2)

    _lookup = dict()

    @classmethod
    def get(cls, descriptor, management, pid=None):
        """ Get IO for a descriptor and pid """
        with management.lock:
            if descriptor not in cls._lookup:
                return

            if not pid:
                return cls._lookup[descriptor]
            else:
                return cls._lookup[descriptor].get(pid, None)

    @classmethod
    def remove(cls, descriptor, management):
        """ Remove IO for a descriptor """
        with management.lock:
            cls._lookup.pop(descriptor, None)
    
    def __init__(self, descriptor, pid, management, io_time=None):
        """ Process related provenance information """
        self.management = management
        self.descriptor = descriptor
        self.pid = pid
        self.times = [[None, None], [None, None]]
        self.operations = [0, 0]

        # Save record to lookup
        with self.management.lock:
            self._lookup.setdefault(descriptor, dict())
            self._lookup[descriptor].setdefault(pid, self)

    def update(self, io_type, op_type, io_time=None):
        """ Update IO record latest times  """
        set_time = io_time if io_time else round(time.time(), 3)

        with self.management.lock:
            # Update times (set start if first operation)
            if not self.times[io_type][self.IO_START]:
                self.times[io_type][self.IO_START] = set_time

            self.times[io_type][self.IO_END] = set_time

            # Add operation
            self.operations[io_type] |= op_type

    def write(self):
        """ Write IO record to the database """
        dir_names = ("read", "write")

        with self.management.lock:
            for direction in range(2):
                query_lookup = ("SELECT start, ops FROM {0} "
                                "WHERE phost = ? AND pstart = ? AND pid = ? AND path = ? AND fcreate = ? ".format(dir_names[direction]))
                query_update = ("REPLACE INTO {0} VALUES (?, ?, ?, ?, ?, ?, ?, ?) ".format(dir_names[direction]))

                file_entry = DescriptorEntry.get(self.descriptor).file_entry
                file_record = FileRecord.get(self.descriptor, self.management)
                process_record = ProcessRecord.get(self.pid, self.management)

                if self.times[direction][self.IO_START] is not None:
                    # Lookup start and operations for previous descriptors between same process and file
                    cursor = self.management.db_connection.cursor()
                    values = (self.management.system_name, process_record.pstart, self.pid, file_entry.paths["abs_real"], file_record.fcreate)
                    cursor.execute(query_lookup, values)
                    result = cursor.fetchone()

                    # Merge values if applicable
                    start = self.times[direction][self.IO_START]
                    end = self.times[direction][self.IO_END]
                    ops = self.operations[direction]
                    if result:
                        start = result[0]
                        ops |= result[1]

                    # Update I/O table
                    values = (self.management.system_name, process_record.pstart, self.pid, file_entry.paths["abs_real"], file_record.fcreate, start, end, ops)
                    cursor.execute(query_update, values)

            self.management.db_connection.commit()



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import json
from repeatfs.provenance.render_base import RenderBase


class RenderJSON(RenderBase):
    """ Render provenance graph as JSON """
    def __init__(self, management):
        self.management = management

    def virt_render(self, process, handle, options=None):
        """ Render JSON """
        graph = self._get_graph(process, options)

        # Convert tuple keys to strings
        str_graph = {}
        for section in graph:
            str_graph[section] = {}
            for key, val in graph[section].items():
                str_key = "|".join(str(kp) for kp in key)
                str_graph[section][str_key] = val

        # Write as JSON
        handle.write(bytes(json.dumps(str_graph).encode("utf8")))



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import functools
import json
import os
import subprocess
import threading
from repeatfs.provenance.process_record import ProcessRecord
from multiprocessing import Queue


class Replication:
    """ Recreate and evaluate provenance replication """
    @classmethod
    def api_receive(cls, api_out):
        """ API: Receive replication related command """
        # Attempt to build replication
        try:
            # Deserialize JSON data and call action
            graph_raw = json.loads(api_out.cmd_info["provenance"])
            provenance = {section: {tuple(k.split("|")): v for (k, v) in graph_raw[section].items()} for section in graph_raw}
            expand = set(tuple(key.split("|")) for key in api_out.cmd_info["expand"])
            replication = Replication(provenance, api_out.core, expand=expand, api_out=api_out)

            getattr(replication, "action_{}".format(api_out.cmd_info["action"]))()

        except json.decoder.JSONDecodeError:
            api_out.respond(status="malformed", message=api_out.cmd_info["provenance"])

        except Exception as e:
            api_out.respond(status="error", message=e)

    def __init__(self, provenance, core, expand=None, api_out=None):
        """ Create replication definition from provenance """
        self.provenance = provenance
        self.core = core
        self.expand = expand if expand else []
        self.api_out = api_out
        self.lock = threading.RLock()

    def _send_stream(self, stream, name):
        """ Send output of stream to client """
        for line in iter(stream.readline, b""):
            extended = {name: line.decode("utf-8")}

            # Ensure output pipe isn't written by both streams simultaneously
            with self.lock:
                self.api_out.respond(status="ok", message="", extended=extended, final=False)

    def is_redirection(self, path):
        """ Check if path is redirection (non-pipe/pty) """
        check_abs = path.startswith(self.core.mount)
        check_rel = not (path.startswith("/") or (":" in path and "/" not in path))

        return check_abs or check_rel

    def _build_command(self, chain):
        """ Build shell command from chain """
        # Connect multiple commands by pipes
        processes = self.provenance["process"]

        commands = []
        for process_id in chain:
            command = processes[process_id]["cmd"][0]
            for arg in processes[process_id]["cmd"][1:]:
                if " " not in arg:
                    command += " {}".format(arg)
                else:
                    if "\"" in arg:
                        command += " '{}'".format(arg)
                    else:
                        command += " \"{}\"".format(arg)
            commands.append(command)

        # Connect multiple commands by pipes
        pipe_cmds = " | ".join(commands)

        # Add file redirection
        for stream in ("stdin", "stdout", "stderr"):
            process = processes[chain[0]] if stream == "stdin" else processes[chain[-1]]
            redirect = process[stream].replace("$$$", self.core.mount)

            # Check for file redirection (TTY and special files fail)
            check_abs = redirect.startswith(self.core.mount)
            check_rel = not (redirect.startswith("/") or ":" in redirect)

            if check_abs or check_rel:
                if stream == "stdin":
                    pipe_cmds += " < {}".format(redirect)
                else:
                    operator = ">" if process["trunc_{}".format(stream)] else ">>"
                    pipe_cmds += " {0} {1}".format(operator, redirect)

        return pipe_cmds

    def _build_chain(self, process_id):
        """ Retrieve full chain of piped commands for any specific command """
        processes = self.provenance["process"]
        chain = []

        # Find left-most member of chain
        left_proc = process_id

        while processes[left_proc]["stdin"].startswith("pipe:"):
            # Note current iteration's process
            current_proc = left_proc

            # Search other processes for other end of pipe
            for search_proc in processes:
                if left_proc == search_proc:
                    continue

                stdio = (processes[search_proc]["stdout"], processes[search_proc]["stderr"])
                if processes[left_proc]["stdin"] in stdio:
                    left_proc = search_proc
                    break

            # Stop if the other end of the pipe wasn't found
            if left_proc == current_proc:
                break

        # Build chain from left-most member
        right_proc = left_proc
        chain.append(right_proc)

        while processes[right_proc]["stdout"].startswith("pipe:") or processes[right_proc]["stderr"].startswith("pipe:"):
            # Note current iteration's process
            current_proc = right_proc

            # Search other processes for other end of pipe
            for search_proc in self.provenance["process"]:
                if right_proc == search_proc:
                    continue

                if not processes[search_proc]["stdin"].startswith("pipe:"):
                    continue

                stdio = (processes[right_proc]["stdout"], processes[right_proc]["stderr"])
                if processes[search_proc]["stdin"] in stdio:
                    right_proc = search_proc
                    break

            # Stop if the other end of the pipe wasn't found
            if right_proc == current_proc:
                break

            # Return hashable tuple
            chain.append(right_proc)

        return tuple(chain)

    def _execute_chain(self, chain):
        """ Execute a process chain """
        # Associate replicated pid with original pid
        def register_pid(orig_proc_id):
            stat_info = ProcessRecord.get_stat_info(os.getpid(), self.core.provenance)
            stat_info["pid"] = os.getpid()

            process_map_mp.put((orig_proc_id, stat_info))
            process_map_mp.close()
            process_map_mp.join_thread()

        process_map_mp = Queue()
        processes = self.provenance["process"]
        run_procs = []

        for process_idx, process_id in enumerate(chain):
            process = processes[process_id]
            stdio_handles = {stdio: subprocess.PIPE for stdio in ("stdin", "stdout", "stderr")}

            # Check CWD
            if process["cwd"]["rel_mount"] is None:
                self.api_out.respond(status="warning", message="Process {} CWD was outside of file system".format(process["pid"]), final=False)
                cwd = self.core.mount
            else:
                cwd = os.path.join(self.core.mount, process["cwd"]["rel_mount"])

            # Redirect stdio
            for stdio in stdio_handles:
                # Setup file redirection
                if self.is_redirection(process[stdio]):
                    if stdio == "stdin":
                        mode = "r"
                    else:
                        mode = "w" if process["trunc_{}".format(stdio)] else "a"

                    stdio_path = process[stdio].replace("$$$", self.core.mount)
                    stdio_handles[stdio] = open(stdio_path, mode)

                # Setup pipes
                elif process[stdio].startswith("pipe:"):
                    if stdio == "stdin":
                        # Check that we don't have an orphan pipe
                        if process_idx > 0:
                            if process["stdin"] == processes[chain[process_idx - 1]]["stdout"]:
                                stdio_handles[stdio] = run_procs[-1].stdout
                            if process["stdin"] == processes[chain[process_idx - 1]]["stderr"]:
                                stdio_handles[stdio] = run_procs[-1].stderr
                    elif stdio == "stderr":
                        # Check for combined stdout/stderr pipe
                        if process["stdout"] == process["stderr"]:
                            stdio_handles[stdio] = subprocess.STDOUT

            args = []
            for arg in process["cmd"]:
                args.append(arg.replace("$$$", self.core.mount))

            preexec = functools.partial(register_pid, process_id)
            run_proc = subprocess.Popen(args, cwd=cwd, shell=False, preexec_fn=preexec, **stdio_handles)
            run_procs.append(run_proc)

            # Show stdout/stderr if not redirected
            for stdio in ("stdout", "stderr"):
                if (stdio_handles[stdio] == subprocess.PIPE) and (not process[stdio].startswith("pipe:")):
                    threading.Thread(target=self._send_stream, args=[getattr(run_proc, stdio), stdio]).start()

        # Wait for completion of all processes
        for run_proc in run_procs:
            run_proc.wait()

        # Close handles
        for run_proc in run_procs:
            for stdio in ("stdin", "stdout", "stderr"):
                if getattr(run_proc, stdio):
                    getattr(run_proc, stdio).close()

        # Convert process map queue to dictionary
        process_map = {}

        for _ in range(len(run_procs)):
            proc_orig_id, proc_repl_info = process_map_mp.get()
            process_map[proc_orig_id] = proc_repl_info

        return process_map

    def _verify_child_match(self, orig_id, repl_id, repl_provenance):
        """ Attempt to match original children to replicated children """
        child_map = []

        # Get children (orig, repl), ordered by time
        proc_ids = (orig_id, repl_id)
        children = [[], []]
        provenance = (self.provenance, repl_provenance)

        for prov_idx in range(len(provenance)):
            for child_id in provenance[prov_idx]["process"]:
                child_process = provenance[prov_idx]["process"][child_id]
                if proc_ids[prov_idx] == (child_process["phost"], str(child_process["parent_start"]), str(child_process["parent_pid"])):
                    children[prov_idx].append(child_id)

            children[prov_idx] = sorted(children[prov_idx])

        # Check for same number of children
        if len(children[0]) != len(children[1]):
            return None

        # Check for same executed commands in order
        for idx in range(len(children[0])):
            child_orig_id = children[0][idx]
            child_repl_id = children[1][idx]
            child_orig_proc = provenance[0]["process"][child_orig_id]
            child_repl_proc = provenance[1]["process"][child_repl_id]

            if child_orig_proc["exe"].split(os.sep)[-1] != child_repl_proc["exe"].split(os.sep)[-1]:
                return None

            child_map.append((child_orig_id, child_repl_proc))

        # All checks passed, return mapping
        return child_map

    def _verify_execution(self, process_map):
        """ Verify replication matched original """
        procs_remaining = list(process_map.items())

        # Find current version of replicated file
        repl_target = self.provenance["file"][tuple(self.provenance["target"][("0",)])]
        repl_path = os.path.join(self.core.root, repl_target["paths"]["rel_mount"])

        with self.core.provenance.lock:
            cursor = self.core.provenance.db_connection.cursor()
            cursor.execute("SELECT * FROM file_last WHERE path = ?", (repl_path, ))
            result = cursor.fetchone()

        if not result:
            self.api_out.respond(status="warning", message="Target file does not appear in replicated provenance, check output for errors", final=False)
            return

        # Build provenance graph for target file
        repl_provenance = self.core.provenance.graph.build_graph(result)

        # Check all processes
        while procs_remaining:
            # Lookup provenance for original and replicated process
            orig_id, repl_info = procs_remaining.pop(0)
            repl_id = (self.core.provenance.system_name, str(repl_info["pstart"]), str(repl_info["pid"]))
            orig_process = self.provenance["process"][orig_id]

            # Verify process ran and was recorded to provenance
            if repl_id not in repl_provenance["process"]:
                self.api_out.respond(status="warning", message="Process {} ({}) does not appear in replicated provenance".format(repl_id[2], " ".join(orig_process["cmd"])), final=False)
                continue

            repl_process = repl_provenance["process"][repl_id]
            repl_cmd = " ".join(repl_process["cmd"])
            self.api_out.respond(status="ok", message="Process {} ({}) executed".format(repl_id[2], repl_cmd), final=False)

            # Verify MD5 checksums match
            if orig_process["hash"] != repl_process["hash"]:
                self.api_out.respond(status="warning", message="Process {} ({}) different version than original".format(repl_id[2], repl_cmd), final=False)

            # Queue children to be verified
            child_map = self._verify_child_match(orig_id, repl_id, repl_provenance)

            if child_map is None:
                self.api_out.respond(status="warning", message="Unable to find children of process {} ({}) ".format(repl_id[2], repl_cmd), final=False)
                continue

            procs_remaining.extend(child_map)

    def get_session_chains(self, filter_expanded):
        """ Get ordered list of process chains for session children """
        session_chains = set()

        # Expand all session leaders and manual requests
        potential_leaders = set(self.provenance["session"].keys())
        expand_remain = set(potential_leaders)
        expand_remain.update(self.expand)
        session_children = set()

        # Find all session children
        while True:
            # Check if expansion is complete
            expand_current = expand_remain.intersection(potential_leaders)
            if len(expand_current) == 0:
                break

            # Find children of current expansion targets
            for process_id, process_info in self.provenance["process"].items():
                parent_id = tuple(str(process_info[key]) for key in ["phost", "parent_start", "parent_pid"])

                if parent_id in expand_current:
                    # Ensure thread leader
                    if (process_info["pstart"], process_info["pid"]) == (process_info["tgid_start"], process_info["tgid"]):
                        # Add process to potential leaders, and note as session child
                        potential_leaders.add(process_id)
                        session_children.add(process_id)

                        # Remove parent process from children if requested
                        if filter_expanded:
                            session_children.discard(parent_id)

                    # Update remaining
                    expand_remain.discard(parent_id)

        # Build chains for all session children
        for process_id in session_children:
            chain = self._build_chain(process_id)
            session_chains.add(chain)

        # Sort processes by start time of first member in chain
        return sorted(session_chains, key=lambda x: self.provenance["process"][x[0]]["pstart"])

    def trace_session_child(self, process_id, session_chains):
        """ Get the session child the process belongs to """
        child_id = process_id

        # Follow tree until root process
        while child_id[2] != "1":
            process = self.provenance["process"][child_id]
            leader = (process["tgid_start"], process["tgid"]) == (process["pstart"], process["pid"])

            if leader:
                # Process is thread group leader
                for chain in session_chains:
                    if child_id in chain:
                        return child_id

                child_id = (process["phost"], str(process["parent_start"]), str(process["parent_pid"]))
            else:
                # Process is thread
                child_id = (process["phost"], str(process["tgid_start"]), str(process["tgid"]))

        return None

    def action_list_cmds(self):
        """ List commands to be executed """
        if self.api_out is None:
            return

        # Obtain processes spawned by session leaders
        session_chains = self.get_session_chains(filter_expanded=True)

        self.api_out.respond(status="ok", message="Command List", final=False)

        for chain in session_chains:
            ids = ", ".join(["|".join(process) for process in chain])
            message = "[{}] {}".format(ids, self._build_command(chain))
            self.api_out.respond(status="ok", message=message, extended={"commands": session_chains}, final=False)

        self.api_out.respond(status="ok", message="", final=True)

    def action_replicate(self):
        """ Execute replication """
        if self.api_out.respond is None:
            return

        self.api_out.respond(status="info", message="Starting replication", final=False)

        # Obtain process chains spawned by session leaders
        session_chains = self.get_session_chains(filter_expanded=True)

        # Execute chains and record process map (orig->repl)
        process_map = {}
        for chain in session_chains:
            process_map.update(self._execute_chain(chain))

        self.api_out.respond(status="info", message="Replication complete", final=False)

        # Verify replication
        self.api_out.respond(status="info", message="Starting verification", final=False)
        self._verify_execution(process_map)
        self.api_out.respond(status="info", message="Verification complete", final=True)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry


class ProcessRecord:
    """ Provides process information """
    UPDATE_THRESHOLD = 20

    _lookup = dict()
    _pipe_cache = dict()

    @classmethod
    def get(cls, pid, management):
        """ Get process entry """
        with management.lock:
            return cls._lookup.get(pid, None)

    @classmethod
    def update(cls, pid, management, ignore_pipes=None):
        """ Create a process entry given a pid, or check for process info updates """
        with management.lock:
            if pid in cls._lookup:
                # Update entry
                process_record = cls._lookup[pid]
                process_record._update(ignore_pipes)
            else:
                # Create and register new entry
                process_record = ProcessRecord(pid, management, ignore_pipes)
                cls._lookup[pid] = process_record

        return process_record

    @classmethod
    def get_stat_info(cls, pid, management):
        """ Get basic process stat info """
        stat_info = {}

        with open("/proc/{0}/stat".format(pid), "r") as handle:
            stat_fields = handle.readline().split(" ")

            # Find end of process name
            for field_mod in range(len(stat_fields) - 1):
                if stat_fields[1 + field_mod].endswith(")"): break

            stat_info["pstart"] = round(int(management.system_boot) + int(stat_fields[21 + field_mod]) / management.hz, 3)
            stat_info["parent_pid"] = int(stat_fields[3 + field_mod])
            stat_info["session_id"] = int(stat_fields[5 + field_mod])

        return stat_info

    def __init__(self, pid, management, ignore_pipes=None):
        """ Process related provenance information """
        self.management = management
        self.pid = pid
        self.cmd = None
        self.cwd_desc = None
        self.trunc_history = set()
        self.count = 0
        self.dirty = True

        # Set remaining fields
        self._update(ignore_pipes=ignore_pipes)

    def _get_cmd(self):
        """ Get current command of process """
        try:
            with open("/proc/{0}/cmdline".format(self.pid), "r") as handle:
                cmd = handle.read()[:-1]

        except PermissionError: cmd = ""

        return cmd

    def _record_pipes(self, ignore_pipes):
        """ Record all pipes connected to standard IO devices """
        for fd in range(3):
            if self.stdio[fd].startswith("pipe:"):
                # Create/retrieve pipe descriptor
                desc_entry = DescriptorEntry.gen_pipe(self.stdio[fd], self.management.core)

                # Manually register pipe provenance (special time=0 indicates pipes)
                self.management.register_open(desc_entry.id, pid=self.pid, record_process=False)
                if fd == 0:
                    self.management.register_read(desc_entry.id, pid=self.pid, io_time=0)
                else:
                    self.management.register_write(desc_entry.id, pid=self.pid, io_time=0)
                self.management.register_close(desc_entry.id, write_process=False)

                # Record in pipe cache
                with self.management.lock:
                    self._pipe_cache.setdefault(self.stdio[fd], set())
                    self._pipe_cache[self.stdio[fd]].add(self.pid)

                if not ignore_pipes:
                    ignore_pipes = set()
                ignore_pipes.add(self.pid)

                # Scan all other processes for this pipe
                for search_pid in os.listdir("/proc"):
                    if not search_pid.isdigit():
                        continue

                    # Skip ignored pipes
                    search_pid = int(search_pid)
                    if search_pid in ignore_pipes:
                        continue

                    for search_fd in range(3):
                        try:
                            search_target = os.readlink("/proc/{0}/fd/{1}".format(search_pid, search_fd))
                        except:
                            continue

                        # If other end of pipe found, register process
                        if search_target == self.stdio[fd]:
                            ProcessRecord.update(search_pid, self.management)

    def _update(self, force=False, ignore_pipes=None):
        """ Update process record with latest information if necessary """
        # Allow repeated checks for updates within threshold
        if not force:
            if self.count > self.UPDATE_THRESHOLD:
                return

            self.count += 1

            if self.cmd == self._get_cmd():
                return

        # Set command
        self.cmd = self._get_cmd()

        # Retrieve stat info
        stat_info = self.get_stat_info(self.pid, self.management)
        for field in stat_info:
            setattr(self, field, stat_info[field])

        # Retrieve thread group ID
        with open("/proc/{0}/status".format(self.pid), "r") as handle:
            for line in handle:
                if line.startswith("Tgid:"):
                    self.tgid = int(line.rstrip().split()[1])
                    break

        # Retrieve parent info
        if self.parent_pid > 0:
            stat_info = self.get_stat_info(self.parent_pid, self.management)
            self.parent_start = stat_info["pstart"]
        else:
            self.parent_start = 0

        # Retrieve tgid leader info
        if self.tgid > 0:
            stat_info = self.get_stat_info(self.tgid, self.management)
            self.tgid_start = stat_info["pstart"]
        else:
            self.tgid_start = 0

        # Retrieve session leader info
        if self.session_id > 0:
            stat_info = self.get_stat_info(self.session_id, self.management)
            self.session_start = stat_info["pstart"]
        else:
            self.session_start = 0

        # Record executable
        try:
            self.exe = os.readlink("/proc/{0}/exe".format(self.pid)) if self.pid > 1 else ""
            try:
                self.md5 = self.management._calculate_hash(self.exe)
            except (PermissionError, FileNotFoundError):
                self.md5 = ""
        except PermissionError:
            self.exe = ""
            self.md5 = ""

        # Record CWD
        try:
            self.cwd = os.readlink("/proc/{0}/cwd".format(self.pid)) if self.pid > 1 else ""
        except PermissionError:
            self.cwd = ""

        # Remove previous CWD descriptor and provenance
        if self.cwd_desc:
            self.management.clean_descriptor(self.cwd_desc.id)
            self.cwd_desc.remove()
            self.cwd_desc = None

        # Update CWD descriptor and provenance (to track CDs)
        if self.cwd:
            cwd_paths = FileEntry.get_paths(self.cwd, self.management.core.root, self.management.core.mount)
            if cwd_paths["orig_type"] == "abs_mount":
                file_entry = FileEntry(cwd_paths["abs_virt"], self.management.core)
                self.cwd_desc = DescriptorEntry(file_entry, None, self.management.core)
                self.management.register_open(self.cwd_desc.id, record_process=False, )
                self.management.register_read(self.cwd_desc.id, self.management.OP_CD, update_process=False, io_time=self.pstart)

        # Record environmental variables
        try:
            self.env = ""
            # with open("/proc/{0}/environ".format(pid), "r") as handle:
            #    self.env = handle.read().rstrip()
        except PermissionError:
            self.env = ""

        # Record stdio connections
        self.stdio = ["", "", ""]
        self.stdio_trunc = [False, False, False]
        for fd in range(3):
            try:
                self.stdio[fd] = os.readlink("/proc/{0}/fd/{1}".format(self.pid, fd)).replace(" (deleted)", "")
                self.stdio_trunc[fd] = (self.stdio[fd] in self.trunc_history)
            except:
                pass
        self.trunc_history.clear()

        # Mark as dirty and add to lookup
        self.dirty = True
        with self.management.lock:
            self._lookup[self.pid] = self

        # Register parent and thread leader process
        if self.parent_pid > 0:
            self.update(self.parent_pid, self.management)

        if self.pid != self.tgid and self.tgid > 0:
            self.update(self.tgid, self.management)

        # Record pipes/processes connected to standard IO devices
        self._record_pipes(ignore_pipes)

        return

    def write(self):
        """ Write process record to the database """
        if not self.dirty:
            return

        with self.management.lock:
            values = (self.management.system_name, self.pstart, self.pid, self.parent_start, self.parent_pid, self.cmd, self.exe,
                      self.md5, self.cwd, self.tgid_start, self.tgid, self.session_start, self.session_id, self.env,
                      self.stdio[0], self.stdio[1], self.stdio[2], self.stdio_trunc[1], self.stdio_trunc[2], self.management.mid)
            cursor = self.management.db_connection.cursor()
            cursor.execute("REPLACE INTO process VALUES ({})".format(",".join(["?"] * 20)), values)
            self.management.db_connection.commit()

        # Write CWD provenance
        if self.cwd_desc:
            self.management.register_close(self.cwd_desc.id, write_process=False)
            self.cwd_desc.remove()
            self.cwd_desc = None

        self.dirty = False

        # Write parent and thread leader entries
        if self.parent_pid > 0:
            self.get(self.parent_pid, self.management).write()

        if self.pid != self.tgid and self.tgid > 0:
            self.get(self.tgid, self.management).write()

        # Write pipe entries
        for fd in range(3):
            if self.stdio[fd].startswith("pipe:"):
                with self.management.lock:
                    # Pipe may be in use twice for this pid (stdout and stderr), check existence and discard
                    if self.stdio[fd] in self._pipe_cache:
                        self._pipe_cache[self.stdio[fd]].discard(self.pid)

                        if len(self._pipe_cache[self.stdio[fd]]) > 0:
                            # Write other end of pipe
                            pipe_pid = next(iter(self._pipe_cache[self.stdio[fd]]))
                            self.get(pipe_pid, self.management).write()
                        else:
                            # Clean up cache once complete
                            del self._pipe_cache[self.stdio[fd]]



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import time
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry


class FileRecord:
    """ Provides file information """
    _lookup = dict()
    _last_cache = dict()
    _dirty_cache = set()

    @classmethod
    def get(cls, descriptor, management):
        """ Get file for a  descriptor """
        with management.lock:
            return cls._lookup.get(descriptor, None)

    @classmethod
    def remove(cls, descriptor, management):
        """ Remove IO for a descriptor """
        with management.lock:
            cls._lookup.pop(descriptor, None)

    @classmethod
    def get_last(cls, file_entry, management):
        """ Get cached latest fcreate of path """
        with management.lock:
            if file_entry.paths["abs_real"] in cls._last_cache:
                return cls._last_cache[file_entry.paths["abs_real"]]

            # Check DB if cache miss
            cursor = management.db_connection.cursor()
            values = (file_entry.paths["abs_real"], )
            cursor.execute("SELECT fcreate FROM file_last WHERE path = ?", values)
            result = cursor.fetchone()

            if result:
                return cls._last_cache.setdefault(file_entry.paths["abs_real"], result["fcreate"])

            # Update cache and DB if first occurrence of file
            return cls.set_last(file_entry, management)

    @classmethod
    def set_last(cls, file_entry, management):
        """ Get cached latest fcreate of path """
        set_time = round(time.time(), 3)

        with management.lock:
            # Update cache
            cls._last_cache[file_entry.paths["abs_real"]] = set_time

            # Update DB
            cursor = management.db_connection.cursor()
            values = (file_entry.paths["abs_real"], set_time)
            cursor.execute("REPLACE INTO file_last VALUES (?, ?)", values)
            management.db_connection.commit()

        return set_time

    def __init__(self, descriptor, management):
        """ Process related provenance information """
        self.management = management
        self.descriptor = descriptor

        # Set file creation time
        file_entry = DescriptorEntry.get(descriptor).file_entry
        self.fcreate = self.get_last(file_entry, self.management)

        # Save record to lookup
        with self.management.lock:
            self._lookup.setdefault(descriptor, self)

    def write(self):
        """ Write IO record to the database """
        file_entry = DescriptorEntry.get(self.descriptor).file_entry

        with self.management.lock:
            if (self.fcreate, file_entry.paths["abs_real"]) in self._dirty_cache:
                return

            cursor = self.management.db_connection.cursor()
            values = (file_entry.paths["abs_real"], self.fcreate, file_entry.file_type)
            cursor.execute("INSERT OR IGNORE INTO file VALUES (?, ?, ?)", values)
            self.management.db_connection.commit()

            self._dirty_cache.add((self.fcreate, file_entry.paths["abs_real"]))



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import io
import os
from collections import deque
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry


class RenderBase:
    """ Render provenance functionality """
    def _get_latest_entry(self, process):
        """ Get latest file entry associated with request process """
        file_path = process.cache_entry.file_entry.derived_source.paths["abs_real"]

        with self.management.lock:
            cursor = self.management.db_connection.cursor()
            cursor.execute("SELECT * FROM file_last WHERE path = ?", (file_path, ))
            result = cursor.fetchone()

        return result

    def _get_graph(self, process, options, op_filter=None):
        """ Build and finalize graph for given process """
        # Lookup active version of requested file
        target_id = self._get_latest_entry(process)

        # Build provenance graph
        return self.management.graph.build_graph(target_id, op_filter=op_filter)

    def virt_render(self, process, handle, options=None):
        handle.write(b"Not implemented\n")

    def render_bytes(self, process, options=None):
        """ Get render as bytes """
        handle = io.BytesIO()
        self.virt_render(process, handle, options)

        return handle.getvalue()



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
import sys
from repeatfs.api import API
from repeatfs.core import Core


def get_api_path(path, configuration):
    """ Build API path and check for validity """
    api_path = os.path.join(path, configuration.values["api"])
    return api_path if os.path.exists(api_path) else None


def replicate(arguments, configuration):
    """ Initiate replication of provenance """
    api_path = get_api_path(arguments.replication_path, configuration)
    if not api_path:
        Core.log("Error: Replication path must be within RepeatFS mount", Core.LOG_OUTPUT)
        return

    # Read provenance data
    with open(arguments.provenance, "r") as handle:
        prov_json = handle.read()

    # List commands mode
    if arguments.list_cmds:
        for result in API.request(api_path, command="replicate", extended={"action": "list_cmds", "provenance": prov_json, "expand": arguments.expand}):
            if result["message"]:
                Core.log(result["message"], Core.LOG_OUTPUT)

        return

    # Setup streams
    stdout = open(arguments.stdout, "w") if arguments.stdout else sys.stdout
    stderr = open(arguments.stderr, "w") if arguments.stderr else sys.stderr

    # Send replicate request to API
    for result in API.request(api_path, command="replicate", extended={"action": "replicate", "provenance": prov_json, "expand": arguments.expand}):
        if result["message"]:
            Core.log("[{}] {}".format(result["status"], result["message"]), Core.LOG_OUTPUT)

        if "stdout" in result:
            Core.log(result["stdout"], Core.LOG_OUTPUT, end="", file=stdout)

        if "stderr" in result:
            Core.log(result["stderr"], Core.LOG_OUTPUT, end="", file=stderr)

    # Close streams
    if stdout != sys.stdout:
        stdout.close()

    if stderr != sys.stderr:
        stderr.close()

def shutdown(path, configuration):
    """ Initiate RepeatFS shutdown """
    api_path = get_api_path(path, configuration)
    if not api_path:
        Core.log("Error: Command must target valid RepeatFS mount", Core.LOG_OUTPUT)
        return

    next(API.request(api_path, command="shutdown"))



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
import re
import stat


class FileEntry:
    @classmethod
    def get_paths(cls, path, root, mount):
        """ Build absolute and relative (to root/mount) paths """
        ret_paths = dict()
        root_term = os.path.join(root, "")
        mount_term = os.path.join(mount, "")

        # Calculate relative path
        ret_paths["relative"] = os.path.join(path, "") #path.rstrip(os.sep)
        ret_paths["orig_type"] = "relative"

        if os.path.isabs(ret_paths["relative"]):
            ret_paths["orig_type"] = "abs_virt"

            # Check for absolute real path
            if ret_paths["relative"].startswith(root_term):
                ret_paths["relative"] = ret_paths["relative"][len(root) + 1:]
                ret_paths["orig_type"] = "abs_real"

            # Check for absolute virtual path
            if ret_paths["relative"].startswith(mount_term):
                ret_paths["relative"] = ret_paths["relative"][len(mount) + 1:]
                ret_paths["orig_type"] = "abs_mount"

        # Set absolute paths
        if ":" in ret_paths["relative"]:
            # Non-disk paths (pipes, future) use relative as absolutes)
            ret_paths["abs_real"] = ret_paths["relative"]
            ret_paths["abs_mount"] = ret_paths["relative"]
            ret_paths["abs_virt"] = ret_paths["relative"]
        else:
            ret_paths["abs_real"] = os.path.join(root_term, ret_paths["relative"])
            ret_paths["abs_mount"] = os.path.join(mount_term, ret_paths["relative"])
            ret_paths["abs_virt"] = os.path.join(os.sep, ret_paths["relative"])

        # Clean paths
        ret_paths["relative"] = ret_paths["relative"].rstrip(os.sep)
        ret_paths["abs_real"] = ret_paths["abs_real"].rstrip(os.sep)
        ret_paths["abs_mount"] = ret_paths["abs_mount"].rstrip(os.sep)
        ret_paths["abs_virt"] = ret_paths["abs_virt"].rstrip(os.sep)

        return ret_paths

    """ Provides meta-data for real and derived files """
    def __init__(self, virt_path, core):
        self.core = core
        self.virt_mtime = 0
        self.init_size = 0
        self.valid = False
        self.api = False
        self.file_type = None
        self.provenance = True
        self.virt_action = None
        self.derived_source = None
        self.derived_actions = dict()

        # Check for inline commands
        inline_sep = self.core.configuration.values["suffix"] * 2
        inline_fields = virt_path.split(inline_sep)
        self.inline_cmd = "".join(inline_fields[1:])

        # Build paths
        virt_path = inline_fields[0]
        self.paths = FileEntry.get_paths(virt_path.lstrip(os.sep), self.core.root, self.core.mount)

        # Check for API
        api_file = os.path.join(os.sep, self.core.configuration.values["api"])

        if self.paths["abs_virt"].endswith(api_file):
            self.file_type = stat.S_IFREG
            self.provenance = False
            self.valid = True
            self.api = True
            return

        # Check for pipes
        if self.paths["abs_virt"].startswith("pipe:"):
            self.file_type = stat.S_IFREG
            self.valid = True
            return

        self._build_entry()

    def __str__(self):
        return "relative {} abs_real {} abs_virt {} type {} action {} derived actions {} valid {}, dervied source:\n{}".format(
            self.paths["relative"], self.paths["abs_real"], self.paths["abs_virt"],
            self.file_type, self.virt_action, self.derived_actions, self.valid, self.derived_source)

    def _build_entry(self):
        """ Calculate details of FileEntry record """
        # Check if a real file or possible derived
        if os.path.lexists(self.paths["abs_real"]):
            # Real path exists
            stats = os.lstat(self.paths["abs_real"])
            self.file_type = stat.S_IFMT(stats.st_mode)
            self.valid = True
            self._populate_time(stats)
            self._populate_actions()
        else:
            virt_dir = os.path.dirname(self.paths["abs_virt"])
            virt_base = os.path.basename(self.paths["abs_virt"])

            # Check for valid derived virtual path
            source_dir = None

            # Must prioritize virtual path as directory, or else nested virtual directories will be recognized as files
            if virt_base.endswith(self.core.configuration.values['suffix']):
                source_dir = self.paths["abs_virt"]
                self.file_type = stat.S_IFDIR
            elif virt_dir.endswith(self.core.configuration.values['suffix']):
                source_dir = virt_dir
                self.file_type = stat.S_IFREG
            else:
                # If neither the full path, nor the parent had a derived suffix, mark invalid
                return

            # If hidden mode and no dot, mark invalid
            source_base = os.path.basename(source_dir)
            if self.core.configuration.values["hidden"] and not source_base.startswith("."):
                return

            # Base directory must have a corresponding file
            source_path = source_dir[:-len(self.core.configuration.values['suffix'])]

            if self.core.configuration.values["hidden"]:
                source_path = os.path.join(os.path.dirname(source_path), os.path.basename(source_path)[1:])

            derived_source = FileEntry(source_path, self.core)
            if not derived_source.valid or (derived_source.derived_source and (derived_source.file_type != stat.S_IFREG)):
                return

            # Consider file successfully derived at this point (allows error handling for new files in derived dirs)
            self.derived_source = derived_source

            if self.file_type == stat.S_IFREG:
                # If potential virtual file, ensure action defined for destination (Dirs are already valid)
                if virt_base not in derived_source.derived_actions:
                    return

                # Set VDF configuration options
                derived_action = derived_source.derived_actions[virt_base][:2]
                self.init_size = self.core.configuration.actions[derived_action]["init_size"]

            self.valid = True
            self._populate_time()
            self._populate_actions()

    def _populate_actions(self):
        """ Populate possible actions for this file """
        virt_base = os.path.basename(self.paths["abs_virt"])

        # Set virtual action if this a derived file
        if (self.file_type == stat.S_IFREG) and self.derived_source:
            self.virt_action = self.derived_source.derived_actions[virt_base]

        for action in self.core.configuration.actions:
            if not self.derived_source or (self.file_type == stat.S_IFREG):
                # For any files or real directories, derived actions are based off file itself (virt_base)
                current_base = virt_base
            else:
                # For derived directories, derived actions are based off target file (target_base)
                current_base = os.path.basename(self.derived_source.paths["abs_virt"])

            # Add derived action if regex matches
            match = re.search(action[0], current_base)
            if match:
                action_name = current_base + action[1]
                self.derived_actions[action_name] = action + (match.groups(), )

    def _populate_time(self, stats=None):
        """ Populate modified time for this file """
        if self.derived_source:
            self.virt_mtime = self.derived_source.virt_mtime
        else:
            self.virt_mtime = stats.st_mtime



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import ast

# Third party modules
import pygraphviz

    def _add_file(self, file_index, graph, info, target):
        """ Add file node to graph """
        file_render = ",".join(map(str, file_index))
        node_color = "green:white" if target else "blue:white"
        graph.add_node(file_index, label=file_index[0], color="black", fillcolor=node_color, style="filled", gradientangle="270", shape="note", URL="javascript:activate_file('{0}');".format(file_render))
        info["file"][file_render] = file_index + ("test", )

    def _add_process(self, entry, graph, info, render):
        """ Add process node to graph """
        process = (entry["phost"], entry["pstart"], entry["pid"])
        parent = [entry["phost"], entry["parent_start"], entry["parent_pid"]]
        process_render = ",".join(map(str, process))
        parent_render = ",".join(map(str, parent))

        # Update info if present
        if len(entry) > 3:
            # Create placeholders for process and parent (to hold children before declared)
            info["process"].setdefault(process_render, [""] * 9 + [[]])
            info["process"].setdefault(parent_render, [""] * 9 + [[]])

            # Update process in info table
            process_start = time.strftime(self.DATE_FORMAT, time.localtime(entry["pstart"]))
            env_render = "<br>".join(sorted(entry["env"].split("\0")[:-1]))
            info["process"][process_render][:9] = [
                entry["phost"], process_start, entry["pid"], parent, entry["cmd"], entry["exe"], entry["hash"], entry["cwd"], env_render]

            # Add process to parent TODO: change this over to a list instead of repeatedly converting
            if list(process) not in info["process"][parent_render][9]:
                info["process"][parent_render][9].append(list(process))

        # If render requested, add to graph
        if render:
            label = info["process"][process_render][4]
            if len(label) > 50:
                label = label[:50] + "..."

            graph.add_node(process, label=label, color="black", fillcolor="red:white", style="filled", gradientangle="270", shape="component", URL="javascript:activate_process('{0}');".format(process_render))

    def _add_io(self, entry, graph, info, write):
        """ Add IO edge to graph """
        process = (entry["phost"], entry["pstart"], entry["pid"])
        process_render = ",".join(map(str, process))
        file_index = (entry["path"], entry["fcreate"])
        file_render = ",".join(map(str, file_index))
        start = time.strftime(self.DATE_FORMAT, time.localtime(entry["start"]))
        stop = time.strftime(self.DATE_FORMAT, time.localtime(entry["stop"]))

        # Direct the graph according to read/write
        if write:
            graph.add_edge(process, file_index, edgeURL="javascript:activate_io('{0}-{1}');".format(process_render, file_render))
            info["io"]["{0}-{1}".format(process_render, file_render)] = ("write", start, stop)
        else:
            graph.add_edge(file_index, process, edgeURL="javascript:activate_io('{0}-{1}');".format(file_render, process_render))
            info["io"]["{0}-{1}".format(file_render, process_render)] = ("read", start, stop)

    def _add_fork(self, parent, child_process, graph, info):
        """ Add forked process IO """
        parent_process = (parent["phost"], parent["pstart"], parent["pid"])
        parent_render = ",".join(map(str, parent_process))
        child_render = ",".join(map(str, child_process))
        fork = time.strftime(self.DATE_FORMAT, time.localtime(child_process[1]))

        # Direct the graph from parent to child
        graph.add_edge(parent_process, child_process, edgeURL="javascript:activate_io('{0}-{1}');".format(parent_render, child_render))
        info["io"]["{0}-{1}".format(parent_render, child_render)] = ("write", fork, fork)

    def _update_processes(self, process, process_tree, info):
        """ Add relatives of process into process tree """
        lineage = list()
        statement = "SELECT * FROM process WHERE AND phost = ? AND pstart = ? AND pid = ?"
        cursor = self.db_connection.cursor()
        cur_process = process

        # Trace lineage
        while cur_process[2] > 0:
            # Retrieve row for current process
            cursor.execute(statement, cur_process)
            row = cursor.fetchone()
            lineage.append((cur_process, row))

            # Iterate to parent
            cur_process = (row["phost"], row["parent_start"], row["parent_pid"])

        # Update tree with lineage, format {proc1: (in_graph, [child_proc1, child_proc2]), proc2, child_proc1, grandchild_proc1, ...}
        parent = (0, 0, 0)
        process_tree.setdefault(parent, (False, set()))

        for cur_process in lineage[::-1]:
            # Add process to process tree
            process_tree.setdefault(cur_process[0], [False, set(), cur_process[1]])
            process_tree[cur_process[0]][0] |= (cur_process[0] == process)
            process_tree[parent][1].add(cur_process[0])
            parent = cur_process[0]

            # Add process to info
            self._add_process(cur_process[1], None, info, False)

    def _build_graph_old(self, path, graph, info, process_tree):
        """ Build graph SVG and sidebar info. Previous is (host, launch, pid, read_stop) """
        # Create queue of remaining paths to process, and add initial path
        remaining = deque()
        remaining.appendleft((path, None))
        io_epsilon = self.core.configuration.values["io_epsilon"]

        with self.lock:
            cursor = self.db_connection.cursor()

            while len(remaining) > 0:
                current, previous = remaining.pop()

                # Create file and writer processes if first time seeing node
                if not graph.has_node(current):
                    # Add file node
                    self._add_file(current, graph, info, previous is None)

                    # Retrieve write/process data (read stop happened after write start)
                    statement = ("SELECT * FROM file NATURAL JOIN write NATURAL JOIN process "
                                 "WHERE path=? AND fcreate = ?")
                    if previous is not None:
                        statement += "AND (write.start = 0 OR write.start <= (? + ?))"
                        cursor.execute(statement, current + (previous[3], io_epsilon))
                    else:
                        cursor.execute(statement, current)

                    for row in cursor:
                        # Start with primary process
                        process_row = dict(row)
                        process = (process_row["phost"], process_row["pstart"], process_row["pid"])
                        # Add code used to be right here

                        # Propagate original time across pipes
                        write_stop = row["stop"]
                        if write_stop == 0:
                            write_stop = previous[3]

                        # Check primary and all parent processes for prior reads
                        while True:
                            read_cursor = self.db_connection.cursor()
                            statement = ("SELECT file.path, file.fcreate, read.stop FROM file NATURAL JOIN read NATURAL JOIN process "
                                         "WHERE phost=? AND pstart=? AND pid=? AND (read.start = 0 OR read.start <= (? + ?)) ")
                            read_cursor.execute(statement, process + (write_stop, io_epsilon))

                            read_row = None
                            for read_row in read_cursor:
                                # Propagate original time across pipes
                                read_stop = read_row["stop"]
                                if read_stop == 0:
                                    read_stop = write_stop

                                # Queue each process read (graphs get too large to perform this recursively)
                                remaining.appendleft(((read_row["path"], read_row["fcreate"]), process + (read_stop, )))

                            # Get current process full info (won't match write's row for parent processes)
                            process_cursor = self.db_connection.cursor()
                            statement = ("SELECT * FROM process "
                                         "WHERE phost=? AND pstart=? AND pid=?")
                            process_cursor.execute(statement, process)
                            process_row = process_cursor.fetchone()

                            # Add process (and write edge) if primary process or participating parent (parent that had read prior to fork)
                            if process == (row["phost"], row["pstart"], row["pid"]):
                                # Primary process and write edge
                                self._add_process(process_row, graph, info, True)
                                self._update_processes(process, process_tree, info)
                                self._add_io(row, graph, info, True)

                                # Note as child participating process for parents
                                child_process = process

                            elif read_row:
                                # Participating parent and fork edge (which read row used doesn't matter)
                                self._add_process(process_row, graph, info, True)
                                self._update_processes(process, process_tree, info)
                                self._add_fork(process_row, child_process, graph, info)

                                # Note as child participating process for parents
                                child_process = process

                            # Setup next parent (read must occur before child process was spawned)
                            if process_row["parent_pid"] == 0: break

                            write_stop = process_row["pstart"]
                            process = (row["phost"], process_row["parent_start"], process_row["parent_pid"])

                # Retrieve read data and connect to previous process (even for previously created nodes)
                if previous is not None:
                    statement = ("SELECT * FROM read "
                                 "WHERE phost=? AND pstart=? AND pid=? AND path=? AND fcreate = ?")
                    cursor.execute(statement, previous[:3] + current)
                    row = cursor.fetchone()
                    self._add_io(row, graph, info, False)

    def _group_process(self, graph, process_tree, info, process):
        """ Group process nodes on a graph with relative process """
        remaining = deque(process_tree[process][1])
        children = list()

        # Display group process node on graph
        self._add_process(process_tree[process][2], graph, info, True)

        # Note all children currently in graph
        while len(remaining) > 0:
            current_process = remaining.pop()
            if process_tree[current_process][0]:
                children.append(current_process)

            remaining.extendleft(process_tree[current_process][1])

        # Move all IO of children to group process, and remove children
        for child in children:
            for edge in graph.edges(child):
                edge0_tuple = ast.literal_eval(edge[0])
                edge1_tuple = ast.literal_eval(edge[1])
                edge0_process = (edge0_tuple == child)

                if edge0_process:
                    # Write (Process->File)
                    process_render = "{0}-{1}".format(",".join(map(str, process)), ",".join(map(str, edge1_tuple)))
                    child_render = "{0}-{1}".format(",".join(map(str, child)), ",".join(map(str, edge1_tuple)))
                    graph.add_edge(process, edge[1], edgeURL="javascript:activate_io('{0}');".format(process_render))
                    graph.delete_edge(child, edge[1])
                else:
                    # Read (File->Process)
                    process_render = "{0}-{1}".format(",".join(map(str, edge0_tuple)), ",".join(map(str, process)))
                    child_render = "{0}-{1}".format(",".join(map(str, edge0_tuple)), ",".join(map(str, child)))
                    graph.add_edge(edge[0], process, edgeURL="javascript:activate_io('{0}');".format(process_render))
                    graph.delete_edge(edge[0], child)

                # Add new IO information
                info["io"]["{0}".format(process_render)] = info["io"]["{0}".format(child_render)]
                del info["io"]["{0}".format(child_render)]

            # Remove old child process
            graph.delete_node(child)

    def _collapse_processes(self, graph, process_tree, info, max_procs):
        """ Collapse by grouping processes by parent processes """
        total_count = 0
        remaining = deque()
        collapse = set()

        # Walk graph until the maximum number of processes has been reached
        remaining.appendleft((0, 0, 0))

        while len(remaining) > 0:
            process = remaining.pop()
            proc_count = 0

            # If the current total plus children is under max, continue.  Otherwise, add to collapse set
            if total_count + len(process_tree[process][1]) < max_procs:
                for child in process_tree[process][1]:
                    # If the child is present in the graph, add to the total count
                    if process_tree[child][0]:
                        proc_count += 1

                total_count += proc_count
                remaining.extendleft(process_tree[process][1])
            else:
                total_count += 1
                collapse.add(process)

        # Group any process nodes that are relatives of members of the collapse set
        for group in collapse:
            self._group_process(graph, process_tree, info, group)

    def _collapse_files(self, path, graph, info):
        """ Collapse by removing internal files """
        files = dict()

        # Iterate through all IO
        for io in graph.edges():
            io0_tuple = ast.literal_eval(io[0])
            io1_tuple = ast.literal_eval(io[1])

            # Check if and where a file node is present
            if len(io0_tuple) == 2:
                file_node, proc_node = io
                file_path = io0_tuple[0]
            elif len(io1_tuple) == 2:
                proc_node, file_node = io
                file_path = io1_tuple[0]
            else:
                # Process to process fork
                continue

            # Note file, process, and IO count
            files.setdefault((file_node, file_path), dict())
            files[(file_node, file_path)].setdefault(proc_node, 0)
            files[(file_node, file_path)][proc_node] += 1

        # Iterate through all files, and remove those with one associated process with both R/W (except requested file)
        for file_info in files:
            # Skip requested file, and files with more than 1 process
            if file_info[1] == path: continue
            if len(files[file_info]) > 1: continue

            # Iterate through the processes (should only be 1)
            for proc_node in files[file_info]:
                # Skip processes that only have one direction of IO
                if files[file_info][proc_node] == 1: continue

                # Remove file (automatically removes IO)
                graph.delete_node(file_info[0])

    def _wrap_graph_head(self, handle, path):
        # Headers
        html = ("<html><head>"
                "<title>Provenance for {0}</title>").format(path).encode()
        html += (b"<style>.sidebar{height:100%;width:450px;background-color:#c4c8d5;position:fixed!important;z-index:1;overflow:hidden}"
                 b".main-title{text-align:center;font-size:26px;padding:20px}"
                 b"table{border-collapse: collapse;}"
                 b"td,th{border: 1px solid black;text-align:left;padding:8px;}"
                 b"tr:nth-child(even){background-color:#DDDDDD} tr:nth-child(odd){background-color:#CCCCCC}</style>"
                 b"</head><body style='padding:0;margin:0;font-family:Arial,sans-serif;'>")

        # Sidebar
        html += (b"<div class='sidebar' style='right:0'>"
                 b"<div id='sidebar-title' style='text-align:center;padding-top:10px'></div>"
                 b"<div id='sidebar-body' style='padding:20px'></div></div>")

        # Start SVG DIV
        html += ("<div style='margin-right:400px'>"
                 "<div class='main-title'>Provenance for {0}</div>"
                 "<div style='text-align:center'>").format(path).encode()

        handle.write(html)

    def _wrap_graph_tail(self, handle, info):
        html = b"</div></div><script>"

        # Intialize info
        html += (b"var file = {};\n"
                 b"var process = {};\n"
                 b"var io = {};\n")

        for node_type in info:
            for item in info[node_type]:
                java_array = "[{0}]".format(str(info[node_type][item])[1:-1])
                html += "{0}['{1}'] = {2};\n".format(node_type, item, java_array).encode()

        # Create sidebar function
        html += (b"function activate_file(id) { "
                 b"document.getElementById('sidebar-title').innerHTML = '<b>File Information</b>';"
                 b"document.getElementById('sidebar-body').innerHTML = '<table style=\"width:100%\">' + "
                 b"'<tr><td>Path</td><td>' + file[id][0] + '</td></tr>' + "
                 b"'</table>';"
                 b"}\n"

                 b"function activate_process(id) { "
                 b"parent_link = \"<a href=\\\"javascript:activate_process('\" + process[id][3] + \"')\\\">\\\"\" + process[process[id][3]][4] + \"\\\"</a><br />\";"
                 b"if (process[id][3][2] == 1) parent_link = \"\\\"\" + process[process[id][3]][4] + \"\\\"\";"
                 b"child_links = '';"
                 b"for (var idx = 0 ; idx < process[id][8].length ; idx++) {"
                 b"   child_links += \"<a href=\\\"javascript:activate_process('\" + process[id][8][idx] + \"')\\\">\\\"\" + process[process[id][8][idx]][4] + \"\\\"</a><br />\";"
                 b"}"
                 b"document.getElementById('sidebar-title').innerHTML = '<b>Process Information</b>';"
                 b"document.getElementById('sidebar-body').innerHTML = '<div style=\"overflow:auto;border: 1px solid black;\"><table style=\"width:100%;border-style:hidden;\">' + "
                 b"'<tr><td>Host Name</td><td>' + process[id][0] + '</td></tr>' + "
                 b"'<tr><td>Start Time</td><td>' + process[id][1] + '</td></tr>' + "
                 b"'<tr><td>Process ID</td><td>' + process[id][2] + '</td></tr>' + "
                 b"'<tr><td>Parent</td><td>' + parent_link + '</td></tr>' + "
                 b"'<tr><td>Command</td><td>' + process[id][4] + '</td></tr>' + "
                 b"'<tr><td>Executable</td><td>' + process[id][5] + '</td></tr>' + "
                 b"'<tr><td>Hash</td><td>' + process[id][6] + '</td></tr>' + "
                 b"'<tr><td>Working Dir</td><td>' + process[id][7] + '</td></tr>' + "
                 b"'<tr><td>Children</td><td>' + child_links + '</td></tr>' + "
                 b"'<tr><td>Environment</td><td>' + process[id][8] + '</td></tr>' + "
                 b"'</table></div>';"
                 b"}\n"

                 b"function activate_io(id) { "
                 b"document.getElementById('sidebar-title').innerHTML = '<b>IO Information</b>';"
                 b"document.getElementById('sidebar-body').innerHTML = '<table style=\"width:100%\">' + "
                 b"'<tr><td>Direction</td><td>' + io[id][0] + '</td></tr>' + "
                 b"'<tr><td>Start</td><td>' + io[id][1] + '</td></tr>' + "
                 b"'<tr><td>End</td><td>' + io[id][2] + '</td></tr>' + "
                 b"'</table>';"
                 b"}\n")

        html += b"</script></body></html>"
        handle.write(html)

    def virt_graph_svg(self, process, handle, collapse):
        file_path = process.cache_entry.file_entry.derived_source.paths["abs_real"]

        # Lookup active version of requested file
        cursor = self.db_connection.cursor()
        cursor.execute("SELECT * FROM file_last WHERE path = ?", (file_path, ))
        result = cursor.fetchone()
        target_file = (result["path"], result["fcreate"])

        # Prepare graph
        graph = pygraphviz.AGraph(directed=True, splines="ortho", ranksep=".7", fontsize="20", pad="0.5")
        graph.node_attr["fontsize"] = "12"

        # Build graph from history
        info = {'file': dict(), 'process': dict(), 'io': dict()}
        process_tree = dict()
        self._build_graph_old(target_file, graph, info, process_tree)

        # Collapse graph if requested
        if collapse:
            self._collapse_processes(graph, process_tree, info, 7)
            self._collapse_files(file_path, graph, info)

        # Save to disk
        graph.layout(prog='dot')
        self._wrap_graph_head(handle, file_path)
        graph.draw(path=handle, format="svg")
        self._wrap_graph_tail(handle, info)

    def repeat_test(self, process, handle):
        """ Create shell script to reproduce results """
        file_path = process.cache_entry.file_entry.derived_soruce.paths["abs_real"]

        # Lookup active version of requested file
        cursor = self.db_connection.cursor()
        cursor.execute("SELECT * FROM file_last WHERE path = ?", (file_path, ))
        result = cursor.fetchone()
        target_file = (result["path"], result["fcreate"])

        # Prepare graph
        graph = pygraphviz.AGraph(directed=True, splines="ortho", ranksep=".7", fontsize="20", pad="0.5")
        graph.node_attr["fontsize"] = "12"

        # Build graph from history
        info = {'file': dict(), 'process': dict(), 'io': dict()}
        process_tree = dict()
        self._build_graph_old(target_file, graph, info, process_tree)

        # Filter and sort visible processes
        visible_procs = []
        process_added = set()

        for pid, proc in process_tree.items():
            # Skip invisible parent processes
            if not proc[0]:
                continue

            # Check if process has already been added
            if pid in process_added:
                continue

            # TODO: remove this, skip run.sh
            if "run.sh" in proc[2]["cmd"]:
                continue

            # Add process to list and record children
            visible_procs.append((proc[2]["pstart"], proc[2]["cwd"], proc[2]["cmd"]))
            process_queue = list(proc[1])

            while len(process_queue) > 0:
                child = process_queue.pop()
                if child in process_added:
                    continue

                process_added.add(child)
                process_queue.extend(list(process_tree[child][1]))

        visible_procs = sorted(visible_procs)

        for proc in sorted(visible_procs):
            handle.write(bytes("cd {} && {}\n".format(*proc[1:]).encode("utf8")))



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import fuse
import io
import json
import os
import threading
from multiprocessing import Process
from repeatfs.provenance.replication import Replication
from repeatfs.cache_entry import CacheEntry
from repeatfs.descriptor_entry import DescriptorEntry


class API:
    """ Provides API information """
    STATE_START, STATE_EXEC = range(2)
    API_BARE, API_SIMPLE, API_FULL = range(3)

    _session_lookup = dict()
    _lock = threading.RLock()

    @classmethod
    def get(cls, desc_id):
        """ Get a session entry """
        with cls._lock:
            return cls._session_lookup.get(desc_id, None)

    @classmethod
    def request(cls, path, command="", extended={}):
        """ Send JSON request to API """
        # Open handle to API
        fd = os.open(path, os.O_RDWR)

        # If command present, send to API
        if command:
            request_data = dict(extended)
            request_data["command"] = command
            request_json = json.dumps(request_data)
            os.write(fd, (request_json + "\n").encode("utf-8"))

        # Read JSON lines and yield next full response
        while True:
            line = b""
            while True:
                char = os.read(fd, 1)
                if char == b"\n":
                    break

                line += char

            json_response = json.loads(line)
            yield json_response

            if json_response["final"]:
                break

    def __init__(self, desc_id, core):
        self.desc_id = desc_id
        self.core = core
        self.state = API.STATE_START
        self.cmd_info = None
        self.exec_lock = threading.RLock()

        # Create input buffer and output pipe descriptors
        self.input = io.StringIO()
        self.output_r, self.output_w = os.pipe()

        # Generate thread-safe session index and register
        with self._lock:
            self._session_lookup[self.desc_id] = self

        # Immediately execute inline commands
        file_entry = DescriptorEntry.get(self.desc_id).file_entry
        if file_entry.inline_cmd:
            self.write("{}\n".format(file_entry.inline_cmd).encode("utf-8"))

    def __str__(self):
        return "id {0} input {1} {2} output {3} {4}".format(
            self.desc_id, self.input[0], self.input[1], self.output[0], self.output[1])

    def get_commands(self):
        """ API command lookup """
        ret_commands = {
            "shutdown": (fuse.fuse_exit, self.API_BARE),
            "config_vdf": (CacheEntry.api_config, self.API_SIMPLE),
            "replicate": (Replication.api_receive, self.API_FULL)}

        return ret_commands

    def remove(self):
        """ Remove descriptor from session lookup """
        with self._lock:
            del self._session_lookup[self.desc_id]

    def read(self, length):
        """ Read data from API """
        # Perform no action if command not yet received
        if self.state == self.STATE_START:
            return ""

        return os.read(self.output_r, length)

    def write(self, buf):
        """ Write data to API """
        cmd_lookup = self.get_commands()

        with self.exec_lock:
            # Perform no action if command already received
            if self.state == self.STATE_EXEC:
                return len(buf)

            # Buffer command and check for complete JSON line
            str_buf = buf.decode("utf8")
            self.input.write(str_buf)

            if "\n" in str_buf:
                self.state = self.STATE_EXEC
                self.input.seek(0)

                try:
                    self.cmd_info = json.loads(self.input.readline())
                    cmd_info = cmd_lookup.get(self.cmd_info["command"], None)
                    if cmd_info:
                        if cmd_info[1] == self.API_BARE:
                            # Immediate API with no arg
                            cmd_info[0]()
                        elif cmd_info[1] == self.API_SIMPLE:
                            # Immediate API with arg
                            cmd_info[0](self)
                        else:
                            # Full API command
                            Process(target=cmd_info[0], args=(self, )).start()
                    else:
                        self.respond(status="unknown", message=self.cmd_info["command"])

                except json.decoder.JSONDecodeError:
                    self.respond(status="malformed", message=self.input.getvalue())

                except Exception as e:
                    self.respond(status="error", message=e)

            return len(buf)

    def respond(self, status="", message="", extended={}, final=True):
        """ Send JSON response to client """
        # Build response dictionary
        response_data = dict(extended)
        response_data.update({"status": status, "message": message, "final": final})

        # Send JSON response line and close pipe if finalized
        os.write(self.output_w, (json.dumps(response_data) + "\n").encode("utf-8"))
        if final:
            os.close(self.output_w)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
from collections import deque
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry


class Graph:
    """ Graph building functionality """
    def __init__(self, management):
        self.management = management

    def _get_graph_id(self, entry, section):
        """ Get requested ID type from row data """
        if section == "fork":
            # Entry is a tuple, first is child ID, second is normal row
            return entry[0] + self._get_graph_id(entry[1], "process")
        else:
            return tuple(str(entry[x]) for x in self.management.db_keys[section])

    def _get_graph_vals(self, entry, section, primary=True):
        """ Get requested values from row data """
        ret_vals = dict()

        if section == "fork":
            # Entry is a tuple, first is child ID, second is normal row
            ret_vals["phost"] = entry[0][0]
            ret_vals["pstart1"] = entry[0][1]
            ret_vals["pid1"] = entry[0][2]
            ret_vals["pstart2"] = entry[1]["pstart"]
            ret_vals["pid2"] = entry[1]["pid"]

        elif section == "session":
            # Session stores process keys
            for name in self.management.db_keys["process"]:
                ret_vals[name] = entry[name]

        else:
            for name, _ in self.management.db_vals[section]:
                ret_vals[name] = entry[name] if name in dict(entry) else None

        # Set graph visibility (no longer used)
        # ret_vals["primary"] = primary

        return ret_vals

    # TODO: add "public" parameter that removes any absolute path info (files, CWDs, environments)
    def _finalize_graph(self, graph):
        """ Finalize graph (calculate common root and rewrite paths) """
        mount_paths = set()
        lineage_fields = ("phost", "parent_start", "parent_pid")
        mount_lookup = self.management._get_mount_lookup()

        # Calcualte common mount path TODO: Migrate all the modifications below, have a token represent common root, rewrite cmd/env/stdio
        for process_id, entry in graph["process"].items():
            root, mount = mount_lookup[entry["mid"]]

            # Only add process from bottom-most session in process tree
            lineage = process_id
            while not lineage[2] == "0":
                if lineage in graph["session"]:
                    mount_paths.add(root)
                    break

                lineage = tuple(str(graph["process"][lineage][field]) for field in lineage_fields)

        common_mount = os.path.commonpath(mount_paths)

        # Update entries with relative paths
        for entry in graph["file"].values():
            entry["paths"] = {"abs_real": entry["path"], "rel_mount": entry["path"][len(common_mount) + 1:]}
            del entry["path"]

        for entry in graph["process"].values():
            # Rewrite CWD
            entry["cwd"] = {"orig": entry["cwd"], "rel_mount": None}
            cwd_paths = FileEntry.get_paths(entry["cwd"]["orig"], root, mount)

            if cwd_paths["orig_type"] == "abs_mount":
                entry["cwd"]["rel_mount"] = cwd_paths["abs_real"][len(common_mount) + 1:]

            # Rewrite absolute paths as relative from mount point
            for field in ("cmd", "env", "stdin", "stdout", "stderr"):
                entry[field] = entry[field].replace(mount, "$$$")

            # Split cmd arguments
            entry["cmd"] = entry["cmd"].split("\0")

    def build_graph(self, target_id, op_filter=None):
        """ Build graph info """
        sections = ("file", "process", "read", "write", "session", "target")
        ret_graph = {key: {} for key in sections}

        # Add provenance target
        target_graph_id = self._get_graph_id(target_id, "file")
        ret_graph["target"] = {(0, ): target_graph_id}

        # Add initial target file to queue
        remaining = deque()
        remaining.appendleft((self._get_graph_id(target_id, "file"), None, None))
        io_epsilon = self.management.core.configuration.values["io_epsilon"]
        if op_filter is None:
            op_filter = self.management.OP_ALL

        with self.management.lock:
            cursor = self.management.db_connection.cursor()

            while len(remaining) > 0:
                file_id, read_process_id, read_stop = remaining.pop()

                # Add file if not already present
                if file_id not in ret_graph["file"]:
                    statement = ("SELECT * FROM file WHERE path=? AND fcreate = ?")
                    cursor.execute(statement, file_id)
                    ret_graph["file"][file_id] = self._get_graph_vals(cursor.fetchone(), "file")

                    # Retrieve write/process data (read stop happened after write start)
                    statement = ("SELECT * FROM file NATURAL JOIN write NATURAL JOIN process "
                                 "WHERE path=? AND fcreate = ? AND (write.ops & ?) > 0 ")

                    if read_stop is not None:
                        statement += "AND (write.start = 0 OR write.start <= (? + ?)) ORDER BY write.start DESC"
                        cursor.execute(statement, file_id + (op_filter, read_stop, io_epsilon))
                    else:
                        statement += " ORDER BY write.start DESC"
                        cursor.execute(statement, file_id + (op_filter, ))
                        
                    for write_row in cursor:
                        # Start with primary process
                        write_process_id = self._get_graph_id(write_row, "process")

                        # Propagate original time across pipes
                        write_stop = write_row["stop"]
                        if write_stop == 0:
                            write_stop = read_stop

                        # Check primary and all parent processes for prior reads
                        lineage_id = write_process_id
                        session_closed = False

                        while True:
                            # Get current process in lineage full info (won't match write's row for parent processes)
                            lineage_cursor = self.management.db_connection.cursor()
                            statement = ("SELECT * FROM process "
                                         "WHERE phost=? AND pstart=? AND pid=?")
                            lineage_cursor.execute(statement, lineage_id)
                            lineage_row = lineage_cursor.fetchone()

                            ret_graph["process"][lineage_id] = self._get_graph_vals(lineage_row, "process")

                            # Record thread group leader
                            thread_id = (lineage_row["phost"], str(lineage_row["tgid_start"]), str(lineage_row["tgid"]))

                            if lineage_id != thread_id:
                                thread_cursor = self.management.db_connection.cursor()
                                thread_cursor.execute(statement, thread_id)
                                thread_row = thread_cursor.fetchone()

                                ret_graph["process"][thread_id] = self._get_graph_vals(thread_row, "process")

                            # Stop tracing back through parents once we hit init (pid 1)
                            if lineage_row["parent_pid"] == 0: break

                            # Close out session if we're at the leader, stop following process reads
                            session_match = (lineage_row["pstart"], lineage_row["pid"]) == (lineage_row["session_start"], lineage_row["session_id"])
                            if not session_closed and session_match:
                                ret_graph["session"][lineage_id] = self._get_graph_vals(lineage_row, "session")
                                session_closed = True

                            if session_closed:
                                lineage_id = (lineage_row["phost"], str(lineage_row["parent_start"]), str(lineage_row["parent_pid"]))
                                continue

                            read_cursor = self.management.db_connection.cursor()
                            statement = ("SELECT file.path, file.fcreate, read.stop FROM file NATURAL JOIN read NATURAL JOIN process "
                                         "WHERE phost=? AND pstart=? AND pid=? AND (read.ops & ?) > 0 AND (read.start = 0 OR read.start <= (? + ?)) "
                                         "ORDER BY read.start DESC ")
                            read_cursor.execute(statement, lineage_id + (op_filter, write_stop, io_epsilon)) # previously write_process_id
                            read_row = None

                            for read_row in read_cursor:
                                # Propagate original time across pipes
                                read_stop = read_row["stop"]
                                if read_stop == 0:
                                    read_stop = write_stop

                                # Queue each process read (graphs get too large to perform this recursively)
                                remaining.appendleft((self._get_graph_id(read_row, "file"), lineage_id, read_stop)) # previously write_process_id

                            # Add process (primary or participating parent)
                            if lineage_id == write_process_id or read_row:
                                if lineage_id == write_process_id:
                                    # Primary with write
                                    ret_graph["write"][self._get_graph_id(write_row, "write")] = self._get_graph_vals(write_row, "write")
                                else:
                                    # Participating parent with fork
                                    pass
                                    #ret_graph["fork"][self._get_graph_id((child_id, lineage_row), "fork")] = self._get_graph_vals((child_id, lineage_row), "fork")

                                # Note as child participating process for parents (was part of fork code)
                                # child_id = lineage_id

                            else:
                                # Non-participating parent
                                ret_graph["process"][lineage_id]["primary"] = False

                            # Setup next parent (read must occur before child process was spawned)
                            write_stop = lineage_row["pstart"]
                            lineage_id = (lineage_row["phost"], str(lineage_row["parent_start"]), str(lineage_row["parent_pid"]))

                # Retrieve read data and connect to previous process (even for previously created nodes)
                if read_process_id is not None:
                    statement = ("SELECT * FROM read "
                                 "WHERE phost=? AND pstart=? AND pid=? AND path=? AND fcreate = ?")
                    cursor.execute(statement, read_process_id + file_id)
                    read_row = cursor.fetchone()
                    ret_graph["read"][self._get_graph_id(read_row, "read")] = self._get_graph_vals(read_row, "read")

        # Finalize graph
        self._finalize_graph(ret_graph)

        return ret_graph



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import hashlib
import os
import sqlite3
import threading
from functools import partial
from repeatfs.descriptor_entry import DescriptorEntry
from repeatfs.file_entry import FileEntry
from repeatfs.provenance.file_record import FileRecord
from repeatfs.provenance.io_record import IORecord
from repeatfs.provenance.process_record import ProcessRecord
from repeatfs.provenance.graph import Graph
from repeatfs.provenance.render_graphviz import RenderGraphviz
from repeatfs.provenance.render_json import RenderJSON


class Management:
    """ Manage provenance information for IO operations """
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
    (OP_IO, OP_ACCESS, OP_CHMOD, OP_CHOWN, OP_ATTR, OP_GETDIR, OP_GETLINK, OP_MKNOD, OP_RMDIR,
     OP_MKDIR, OP_STATS, OP_UNLINK, OP_MKSYM, OP_MKHARD, OP_MOVE, OP_TIME, OP_CD, OP_TRUNCATE) = [2**x for x in range(18)]
    OP_ALL = 2**19 - 1

    def __init__(self, core):
        self.core = core
        self.enable = True
        self.lock = threading.RLock()

        # Setup graphing and renderers
        self.graph = Graph(self)
        self.render_json = RenderJSON(self)
        self.render_graphviz = RenderGraphviz(self)

        # Cache system values
        self.system_name = os.uname().nodename
        self.system_boot = self._get_boot()
        self.hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

        # Make DB connection (creates DB if doesn't exist)
        db_path = os.path.join(core.configuration.path, "provenance.db")
        self.db_connection = sqlite3.connect(db_path, check_same_thread=False)
        self._init_db()

        # Register and refresh targets
        self._write_root()

    def _init_db(self):
        """ Create DB structure """
        self.db_keys = dict()
        self.db_keys["mount"] = ("mid",)
        self.db_keys["file"] = ("path", "fcreate")
        self.db_keys["file_last"] = ("path", )
        self.db_keys["process"] = ("phost", "pstart", "pid")
        self.db_keys["read"] = ("phost", "pstart", "pid", "path", "fcreate")
        self.db_keys["write"] = ("phost", "pstart", "pid", "path", "fcreate")

        self.db_vals = dict()
        self.db_vals["mount"] = [("mid", "integer"), ("root", "text"), ("mount", "text")]
        self.db_vals["file"] = [("path", "text"), ("fcreate", "int"), ("type", "text")]
        self.db_vals["file_last"] = [("path", "text"), ("fcreate", "int")]
        self.db_vals["process"] = [("phost", "text"), ("pstart", "int"), ("pid", "int"), ("parent_start", "int"), ("parent_pid", "int"),
                                   ("cmd", "blob"), ("exe", "text"), ("hash", "text"), ("cwd", "text"),
                                   ("tgid_start", "int"), ("tgid", "int"), ("session_start", "int"), ("session_id", "int"), ("env", "text"), 
                                   ("stdin", "text"), ("stdout", "text"), ("stderr", "text"),
                                   ("trunc_stdout", "int"), ("trunc_stderr", "int"), ("mid", "int")]
        self.db_vals["read"] = [("phost", "text"), ("pstart", "int"), ("pid", "int"), ("path", "text"), ("fcreate", "int"),
                                ("start", "int"), ("stop", "int"), ("ops", "int")]
        self.db_vals["write"] = [("phost", "text"), ("pstart", "int"), ("pid", "int"), ("path", "text"), ("fcreate", "int"),
                                 ("start", "int"), ("stop", "int"), ("ops", "int")]

        self.db_ddl = list()
        self.db_ddl.append("CREATE UNIQUE INDEX IF NOT EXISTS mount_rootmount ON mount(root, mount)")
        self.db_ddl.append("CREATE INDEX IF NOT EXISTS process_parent ON process(phost, parent_start, parent_pid)")

        # Create queries
        tables = dict()
        for table in self.db_vals:
            tables[table] = ",".join([" ".join(x) for x in self.db_vals[table]])
            tables[table] += ", PRIMARY KEY ({})".format(",".join(self.db_keys[table]))

        with self.lock:
            self.db_connection.row_factory = sqlite3.Row
            cursor = self.db_connection.cursor()

            # Currently disable synchronous mode for performance
            cursor.execute("PRAGMA synchronous = OFF")

            # Create tables
            for table in tables:
                cursor.execute("CREATE TABLE IF NOT EXISTS {0} ({1})".format(table, tables[table]))

            # Run manually specified DDL queries
            for ddl in self.db_ddl:
                cursor.execute(ddl)

            # Save changes
            self.db_connection.commit()

    def _get_boot(self):
        """ Get system boot time """
        with open("/proc/stat", "r") as handle:
            for line in handle:
                if line.startswith("btime"):
                    return float(line.split(" ")[1].rstrip())

    def _calculate_hash(self, path):
        """ Generate MD5 hash for file """
        md5 = hashlib.md5()

        with open(path, "rb") as handle:
            for chunk in iter(partial(handle.read, 4096), b""):
                md5.update(chunk)

        return md5.hexdigest()

    def _write_root(self):
        """ Write current FS root and refresh IDs """
        with self.lock:
            # Register current FS root if necessary
            cursor = self.db_connection.cursor()
            cursor.execute("INSERT OR IGNORE INTO mount VALUES (NULL, ?, ?)", (self.core.root, self.core.mount))
            self.db_connection.commit()

            # Populate root ID lookup
            cursor.execute("SELECT * FROM mount ORDER BY root=? AND mount=? DESC", (self.core.root, self.core.mount))
            self.mid = cursor.fetchone()["mid"]

    def _get_mount_lookup(self):
        """ Get mount lookup table """
        cursor = self.db_connection.cursor()
        cursor.execute("SELECT * FROM mount")

        return {row["mid"]: (row["root"], row["mount"]) for row in cursor}

    # TODO: write and update_last are separate because of pipes, but pipes should eventually be recorded in last time, and params combined to "create"
    def register_open(self, descriptor, pid=None, read=False, write=False, record_file=True, record_process=True, update_last=False):
        """ Cache process, file, and IO """
        if not self.enable:
            return

        pid = self.core.get_pid(pid)

        # Start new version of file
        if update_last:
            file_entry = DescriptorEntry.get(descriptor).file_entry
            FileRecord.set_last(file_entry, self)

        # Record file
        if record_file:
            FileRecord(descriptor, self)

        # Record IO 
        IORecord(descriptor, pid, self)

        # Record process
        if record_process:
            ProcessRecord.update(pid, self)

        # Record initial read/write
        if read:
            self.register_read(descriptor, op_type=self.OP_IO, pid=pid, update_process=False)

        if write:
            self.register_write(descriptor, op_type=(self.OP_IO | self.OP_TRUNCATE), pid=pid, update_process=False)

    def register_close(self, descriptor, write_process=True):
        """ Write IO records if IO occurred """
        if not self.enable:
            return

        with self.lock:
            # Write file record
            FileRecord.get(descriptor, self).write()

            # Write each process and IO record for this descriptor
            for pid in IORecord.get(descriptor, self):
                io_record = IORecord.get(descriptor, self, pid=pid)
                io_record.write()

                if write_process:
                    ProcessRecord.get(pid, self).write()

        # Remove descriptor from caches
        self.clean_descriptor(descriptor)

    def register_read(self, descriptor, op_type=OP_IO, pid=None, update_process=True, io_time=None):
        """ Update read end time """
        if not self.enable:
            return

        pid = self.core.get_pid(pid)

        # Ensure pid recorded to this descriptor (for descriptors passed to child processes)
        if pid not in IORecord.get(descriptor, self):
            self.register_open(descriptor, pid=pid, record_file=False)

        IORecord.get(descriptor, self, pid=pid).update(IORecord.IO_READ, op_type, io_time=io_time)

        if update_process:
            ProcessRecord.get(pid, self)._update()

    def register_write(self, descriptor, op_type=OP_IO, pid=None, update_process=True, io_time=None):
        """ Update write end time """
        if not self.enable:
            return

        pid = self.core.get_pid(pid)

        # Ensure pid recorded to this descriptor (for descriptors passed to child processes)
        if pid not in IORecord.get(descriptor, self):
            self.register_open(descriptor, pid=pid, record_file=False)

        IORecord.get(descriptor, self, pid=pid).update(IORecord.IO_WRITE, op_type, io_time=io_time)

        # Note truncations in process record for potential file redirection
        if op_type & self.OP_TRUNCATE:
            file_entry = DescriptorEntry.get(descriptor).file_entry
            ProcessRecord.get(pid, self).trunc_history.add(file_entry.paths["abs_mount"])

        if update_process:
            ProcessRecord.get(pid, self)._update()

    def register_op_read(self, file_entry, op_type):
        """ Register an ephemeral read operation """
        if not self.enable:
            return

        with DescriptorEntry(file_entry, None, self.core) as desc_entry:
            self.register_open(desc_entry.id)
            self.register_read(desc_entry.id, op_type)
            self.register_close(desc_entry.id)

    def register_op_write(self, file_entry, op_type, create=False):
        """ Register an ephemeral write operation """
        if not self.enable:
            return

        with DescriptorEntry(file_entry, None, self.core) as desc_entry:
            self.register_open(desc_entry.id, update_last=create)
            self.register_write(desc_entry.id, op_type)
            self.register_close(desc_entry.id)

    def clean_descriptor(self, descriptor):
        """ Remove descriptor based records (file and IO) """
        FileRecord.remove(descriptor, self)
        IORecord.remove(descriptor, self)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import fuse as fuse


class Fuse(fuse.Operations):
    """ Implements FUSE API methods """
    def __init__(self, core):
        self.core = core
        self.fuse_error = fuse.FuseOSError

    def fuse_get_context(self):
        """ Provide fuse_get_context info externally """
        return fuse.fuse_get_context()

    def access(self, path, mode):
        self.core.log("CALL: access ({0}, {1})".format(path, mode), self.core.LOG_CALL)
        return self.core.get_access(path, mode)

    def chmod(self, path, mode):
        self.core.log("CALL: chmod ({0}, {1})".format(path, mode), self.core.LOG_CALL)
        return self.core.change_mode(path, mode)

    def chown(self, path, uid, gid):
        self.core.log("CALL: chown ({0}, {1}, {2})".format(path, uid, gid), self.core.LOG_CALL)
        return self.core.change_owner(path, uid, gid)

    def getattr(self, path, info=None):
        self.core.log("CALL: getattr ({0}) ({1})".format(path, info.fh if info else None), self.core.LOG_CALL)
        return self.core.get_attributes(path, info)

    def opendir(self, path):
        self.core.log("CALL: opendir ({0})".format(path), self.core.LOG_CALL)
        return self.core.open_directory(path)

    def readdir(self, path, fh):
        self.core.log("CALL: readdir ({0}, {1})".format(path, fh), self.core.LOG_CALL)
        return self.core.get_directory(path, fh)

    def readlink(self, path):
        self.core.log("CALL: readlink ({0})".format(path), self.core.LOG_CALL)
        return self.core.get_link(path)

    def mknod(self, path, mode, dev):
        self.core.log("CALL: mknod ({0}, {1}, {2})".format(path, mode, dev), self.core.LOG_CALL)
        return self.core.create_node(path, mode, dev)

    def rmdir(self, path):
        self.core.log("CALL: rmdir ({0})".format(path), self.core.LOG_CALL)
        return self.core.remove_directory(path)

    def mkdir(self, path, mode):
        self.core.log("CALL: mkdir ({0}, {1})".format(path, mode), self.core.LOG_CALL)
        return self.core.create_directory(path, mode)

    def statfs(self, path):
        self.core.log("CALL: statfs ({0})".format(path), self.core.LOG_CALL)
        return self.core.fs_stats(path)

    def unlink(self, path):
        self.core.log("CALL: unlink ({0})".format(path), self.core.LOG_CALL)
        return self.core.unlink(path)

    def symlink(self, name, target):
        self.core.log("CALL: symlink ({0}, {1})".format(name, target), self.core.LOG_CALL)
        return self.core.make_symlink(target, name)

    def link(self, name, target):
        self.core.log("CALL: link ({0}, {1})".format(target, name), self.core.LOG_CALL)
        return self.core.make_hardlink(target, name)

    def rename(self, old, new):
        self.core.log("CALL: rename ({0}, {1})".format(old, new), self.core.LOG_CALL)
        return self.core.rename(old, new)

    def utimens(self, path, times=None):
        self.core.log("CALL: utimens ({0}, {1})".format(path, times), self.core.LOG_CALL)
        return self.core.update_time(path, times)

    def open(self, path, info):
        self.core.log("CALL: open ({0}, {1})".format(path, info.flags), self.core.LOG_CALL)
        return self.core.open(path, info)

    def create(self, path, mode, info):
        self.core.log("CALL: create ({0}, {1})".format(path, mode), self.core.LOG_CALL)
        return self.core.open(path, info, mode=mode)

    def read(self, path, length, offset, info=None):
        self.core.log("IO: read ({0}, {1}, {2}, {3})".format(path, length, offset, info.fh), self.core.LOG_IO)
        return self.core.read(path, length, offset, info)

    def write(self, path, buf, offset, info=None):
        self.core.log("IO: write ({0}, {1}, {2}, {3})".format(path, offset, info.fh, info.writepage), self.core.LOG_IO)
        return self.core.write(path, buf, offset, info)

    def truncate(self, path, length, info=None):
        self.core.log("CALL: truncate ({0}, {1})".format(path, length), self.core.LOG_CALL)
        return self.core.truncate(path, length, info)

    # TODO: Verify this sync
    def flush(self, path, info):
        self.core.log("CALL: flush ({0}, {1})".format(path, info.fh), self.core.LOG_CALL)
        return self.core.sync(path, info)
        # return 0

    def release(self, path, info):
        self.core.log("CALL: release ({0}, {1})".format(path, info.fh), self.core.LOG_CALL)
        return self.core.close(info)

    def fsync(self, path, fdatasync, info):
        self.core.log("CALL: fsync ({0}, {1}, {2})".format(path, fdatasync, info.fh), self.core.LOG_CALL)
        return self.core.sync(path, info)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import io
import json
import os
from repeatfs.provenance.replication import Replication
from repeatfs.provenance.render_base import RenderBase

# Third party modules
import pygraphviz


class RenderGraphviz(RenderBase):
    """ Render provenance graph as Graphviz SVG """
    STYLES = {
        "process": {"color": "black", "fillcolor": "red:white", "style": "filled", "gradientangle": "270", "shape": "component", "fontsize": "18"},
        "file": {"color": "black", "fillcolor": "blue:white", "style": "filled", "gradientangle": "270", "shape": "note", "fontsize": "18"},
        "target": {"color": "black", "fillcolor": "green:white", "style": "filled", "gradientangle": "270", "shape": "note", "fontsize": "18"},
        "io": {"color": "black"},
        "fork": {"color": "red"}
    }

    def __init__(self, management):
        self.management = management

    def _build_elements(self, graph, expand_procs, collapse_files):
        """ Build elements given session chains """
        # Note target file
        target = graph["target"][(0,)]
        if collapse_files:
            target = (graph["file"][target]["paths"]["abs_real"], )

        # Get session children processes
        replication = Replication(graph, self.management.core, expand=expand_procs)
        session_chains = replication.get_session_chains(filter_expanded=False)
        session_children = set(child for chain in session_chains for child in chain)

        # Note all processes associated with each IO type per file
        file_groups = {}

        for io_type in ("read", "write"):
            for io_id in graph[io_type]:
                io_file = io_id[3:]
                io_process = replication.trace_session_child(io_id[:3], session_chains)

                # To collapse multiple versions of file, store by path only
                if collapse_files:
                    io_file = (graph["file"][io_file]["paths"]["abs_real"], )

                file_groups.setdefault(io_file, {"read": dict(), "write": dict(), "fcreate": set()})
                file_groups[io_file][io_type][io_process] = io_id;
                file_groups[io_file]["fcreate"].add(io_id[4])

        # Build IO groups (common read, common write pairings for each file)
        io_groups = {}

        for file_id in file_groups:
            # Freeze sets to allow hashing
            read_group = frozenset(file_groups[file_id]["read"].keys())
            write_group = frozenset(file_groups[file_id]["write"].keys())

            if file_id == target:
                # Ensure target is always in its own IO group
                io_group = (read_group, write_group, file_id)
            else:
                io_group = (read_group, write_group)

            # Note file and original IO associated with each IO group
            io_groups.setdefault(io_group, {"files": {}, "target": file_id == target, "read_lookup": {}, "write_lookup": {}})
            io_groups[io_group]["files"][file_id] = file_groups[file_id]["fcreate"]

            for io_type in ("read", "write"):
                for io_id in file_groups[file_id][io_type]:
                    io_groups[io_group]["{}_lookup".format(io_type)].setdefault(io_id, [])
                    io_groups[io_group]["{}_lookup".format(io_type)][io_id].append(file_groups[file_id][io_type][io_id])

        return {"process": session_children, "io": io_groups}

    def _add_process(self, process_id, graph, svg):
        """ Add process to SVG/JS """
        # Add SVG
        label = " ".join(graph["process"][process_id]["cmd"])
        label = label.replace("/home/anthonyw", "/opt")
        if len(label) > 40:
            label = "{}...".format(label[:40])

        if not svg.has_node(process_id):
            svg.add_node(process_id, label=label, URL="javascript:activate_process('{}');".format("|".join(process_id)), **self.STYLES["process"])

        # Add Javascript

    def _add_io_group(self, group_id, files_info, graph, svg):
        """ Add IO group to SVG/JS """
        # Build IDs and label
        node_id = []
        for path_id in files_info["files"]:
            for fcreate in files_info["files"][path_id]:
                node_id.append((path_id[0], fcreate))

        js_files = ["'{}'".format("|".join(file_id)) for file_id in node_id]

        if len(node_id) == 1:
            label = os.path.basename(node_id[0][0])
        else:
            label = "{} files".format(len(node_id))

        # Add group SVG
        if not svg.has_node(node_id):
            style = "target" if files_info["target"] else "file"
            svg.add_node(node_id, label=label, URL="javascript:activate_file([{}]);".format(",".join(js_files)), **self.STYLES[style])

        # Add IO SVG (read, write)
        for io_idx, io_type in enumerate(("read", "write")):
            if len(group_id[io_idx]) > 0:
                for process_id in group_id[io_idx]:
                    edge = (node_id, process_id) if io_idx == 0 else (process_id, node_id)

                    js_io = []
                    for orig_id in files_info["{}_lookup".format(io_type)][process_id]:
                        js_io.append("'{}'".format("|".join(orig_id)))

                    if not svg.has_edge(*edge):
                        svg.add_edge(*edge, URL="javascript:activate_io({}, [{}]);".format(io_idx, ",".join(js_io)))

    def _add_fork(self, parent_id, child_id, svg):
        """ Add fork between processes """
        edge = (parent_id, child_id)
        if not svg.has_edge(*edge):
            svg.add_edge(*edge, URL="javascript:activate_fork('{}', '{}');".format("|".join(parent_id), "|".join(child_id)), **self.STYLES["fork"])

    def _add_elements(self, graph, svg, expand_procs=None, collapse_files=True):
        """ Add elements from graph to SVG/JS """
        # Build elements for session chains
        elements = self._build_elements(graph, expand_procs, collapse_files)

        # Render processes
        for process_id in elements["process"]:
            self._add_process(process_id, graph, svg)

        # Render IO groups
        for group_id in elements["io"]:
            self._add_io_group(group_id, elements["io"][group_id], graph, svg)

        # Render forks
        for process_id in elements["process"]:
            process_info = graph["process"][process_id]
            parent_id = (process_info["phost"], str(process_info["parent_start"]), str(process_info["parent_pid"]))

            if parent_id in elements["process"]:
                self._add_fork(parent_id, process_id, svg)

    def virt_render(self, process, handle, options=None):
        """ Render SVG """
        op_filter = self.management.OP_IO | self.management.OP_MOVE  # | self.management.OP_ALL
        graph = self._get_graph(process, options, op_filter=op_filter)

        # Prepare graph
        svg = pygraphviz.AGraph(directed=True, splines="ortho", ranksep=".7", fontsize="20", pad="0.5")
        svg.node_attr["fontsize"] = "12"
        options = {}
        options["expand_procs"] = process.cache_entry.config.get("expand_procs", [])

        self._add_elements(graph, svg, **options)

        svg.layout(prog='dot')
        svg_handle = io.BytesIO()
        svg.draw(path=svg_handle, format="svg")

        # Load template
        fields = {}
        fields["python-target"] = os.path.basename(graph["file"][graph["target"][(0,)]]["paths"]["abs_real"]).encode("utf-8")
        fields["python-path"] = process.cache_entry.file_entry.paths["abs_real"].replace("/", "|").encode("utf-8")
        fields["python-svg"] = svg_handle.getvalue()
        fields["python-graph"] =  self.management.render_json.render_bytes(process).replace(b"\"", b"\\\"")
        fields["python-expanded"] = json.dumps(["|".join(val) for val in options["expand_procs"]]).encode("utf-8").replace(b"\"", b"\\\"")
        fields["python-api"] = self.management.core.configuration.values["api"].encode("utf-8")
        fields["python-suffix"] = (self.management.core.configuration.values["suffix"] * 2).encode("utf-8")
        fields["python-session"] = process.cache_entry.config.get("session", "0").encode("utf-8")

        template_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "render_graphviz.html")

        with open(template_path, "rb") as template_handle:
            template = template_handle.read()

        for field in fields:
            template = template.replace("[{}]".format(field).encode("utf-8"), fields[field])

        # Write to process handle
        handle.write(template)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
import re
import sys


class Configuration:
    """ Storages global and per-file configurations """

    # Casts must be defined before CONFIG_FIELDS
    def cast_bool(val):
        """ Boolean cast  """
        return val.lower() == "true"

    def cast_path(val):
        """ Path cast """
        return os.path.expanduser(val)

    def cast_list(val):
        """ List cast (if applicable) """
        if isinstance(val, list):
            return val
        else:
            return [val]

    FIELD_MODE = 0
    FIELD_REQ = 1
    FIELD_DEF = 2
    FIELD_TYPE = 3

    TEMPLATE_EXAMPLES = ("# FASTQ -> FASTA\n"
                         "[entry]\n"
                         "match=\.fastq$\n"
                         "ext=.fasta\n"
                         "cmd=seqtk seq -A {input}\n")

    CONFIG_FILE = "repeatfs.conf"

    CONFIG_FIELDS = {
        "suffix": (False, False, "+", str, "virtual directory suffix"),
        "hidden": (False, False, "False", cast_bool, "prepend '.' to virtual directory paths"),
        "invisible": (False, False, "True", cast_bool, "hide virtual directories from directory listing"),
        "block_size": (False, False, "1048576", int, "filesystem block size"),
        "store_size": (False, False, "1073741824", int, "total filestore size"),
        "read_timeout": (False, False, "1.0", float, "read timeout (seconds)"),
        "cache_path": (False, False, "/tmp/repeatfs.cache", cast_path, "cache path"),
        "io_epsilon": (False, False, "7.0", float, "provenance IO is considered simultaneous within this epsilon (seconds)"),
        "api": (False, False, ".repeatfs-api", str, "file for RepeatFS API and control"),
        "api_size": (False, False, "1048576", int, "reported size of RepeatFS API and control"),
        "match": (True, True, None, str, ""),
        "ext": (True, True, None, str, ""),
        "cmd": (True, True, None, str, ""),
        "output": (True, False, "stdout", str, ""),
        "append": (True, False, None, str, ""),
        "disk_cache": (True, False, "True", cast_bool, ""),
        "init_size": (True, False, "0", int, ""),
        "internal": (True, False, None, cast_list, "")}

    SYSTEM_ENTRIES = [
            {"match": ".*", "ext": ".provenance.html", "cmd": "cat {temp}", "internal": ["self.cache_entry.core.provenance.render_graphviz.virt_render", {"collapse": False}], "init_size": "10485760"},
            {"match": ".*", "ext": ".provenance.json", "cmd": "cat {temp}", "internal": ["self.cache_entry.core.provenance.render_json.virt_render"]}]

    @classmethod
    def write_template(cls, path):
        """ Write template configuration to config directory """
        config_path = os.path.join(path, cls.CONFIG_FILE)

        # Do not overwrite an existing file
        if os.path.isfile(config_path):
            print("Error: configuration file already exists")
            return

        # Create directory and template
        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, cls.CONFIG_FILE), "w") as handle:
            handle.write("# Configuration Template\n\n")

            for field in cls.CONFIG_FIELDS:
                if not cls.CONFIG_FIELDS[field][0]:
                    handle.write("## {0}\n#{1}={2}\n\n".format(cls.CONFIG_FIELDS[field][4], field, cls.CONFIG_FIELDS[field][2]))

            handle.write("\n{0}".format(cls.TEMPLATE_EXAMPLES))

        print("Configuration created: {0}".format(config_path))

    def __init__(self, core, path):
        self.values = dict()
        self.actions = dict()
        self.core = core
        self.path = path

        if not self.read_config(os.path.join(path, self.CONFIG_FILE)):
            sys.exit(1)

    def _add_entry(self, entry_mode, values):
        # Check for required fields and set default values
        for field in self.CONFIG_FIELDS:
            if self.CONFIG_FIELDS[field][self.FIELD_MODE] == entry_mode:
                if self.CONFIG_FIELDS[field][self.FIELD_REQ] and field not in values:
                    return "required field '{0}' missing".format(field)

                if not self.CONFIG_FIELDS[field][self.FIELD_REQ] and field not in values:
                    values[field] = self.CONFIG_FIELDS[field][self.FIELD_DEF]

        # Check for conflicts
        if "output" in values and "cmd" in values:
            if values["output"] != "file" and "{output}" in values["cmd"]:
                return "'{output}' command variable only valid for 'file' output"

        # Make active
        for field in values:
            value = None
            if values[field] is not None:
                value = Configuration.CONFIG_FIELDS[field][Configuration.FIELD_TYPE](values[field])

            if entry_mode:
                entry_key = (values['match'], values['ext'])
                self.actions.setdefault(entry_key, dict())
                self.actions[entry_key][field] = value
            else:
                self.values[field] = Configuration.CONFIG_FIELDS[field][Configuration.FIELD_TYPE](values[field])

        return None

    # Read configuration
    def read_config(self, path):
        values = {}

        invalid = None

        try:
            with open(path, "r") as handle:
                entry_mode = False
                line_num = 0

                for line in handle:
                    line = line.rstrip()
                    line_num += 1

                    # Skip comments
                    if re.match("^[ \t]*(#.*)*$", line): continue

                    # Process entry header
                    if re.match("^[ \t]*\[entry\][ \t]*(#.*)*$", line):
                        # Complete the previous entry
                        invalid = self._add_entry(entry_mode, values)
                        if invalid: break

                        # Start new entry
                        values = {}
                        entry_mode = True
                        continue

                    # Check field validity
                    match = re.search("^[ \t]*([^= \t]+)[ \t]*=[ \t]*([^#]+)(#.*)*$", line)

                    if not match or match.groups(1)[0] not in Configuration.CONFIG_FIELDS:
                        invalid = "Invalid line in configuration"
                        break
                    if entry_mode and not Configuration.CONFIG_FIELDS[match.groups(1)[0]][Configuration.FIELD_MODE]:
                        invalid = "Global attribute in entry section"
                        break
                    if not entry_mode and Configuration.CONFIG_FIELDS[match.groups(1)[0]][Configuration.FIELD_MODE]:
                        invalid = "Entry attribute in global section"
                        break

                    # Save field value
                    values[match.groups(1)[0]] = match.groups(1)[1]

            # Add final entry
            if not invalid:
                invalid = self._add_entry(entry_mode, values)

            # Check for errors
            if invalid:
                print("Configuration error: {0} ({1})".format(invalid, line_num))
                return False

            # Add system entries
            for entry in Configuration.SYSTEM_ENTRIES:
                self._add_entry(True, entry)

            return True

        except IOError:
            print("Error: configuration file not found at location `{}`.  Please run 'repeatfs generate' to create a new configuration.".format(path))
            return False



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='RepeatFS',
    version='0.10.0',
    author='Anthony Westbrook',
    author_email='anthony.westbrook@unh.edu',
    packages=setuptools.find_packages(),
    scripts=['scripts/repeatfs'],
    include_package_data=True,
    url='http://github.com/ToniWestbrook/repeatfs',
    license='LICENSE',
    description='File system providing reproducibility through provenance and automation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        "fusepy >= 3.0.1",
        "python-daemon >= 2.2.0",
        "pygraphviz >= 1.3.1",
    ],
)



#! /usr/bin/env python3

#
#   This file is part of RepeatFS
#
#   SPDX-FileCopyrightText: 2020  Anthony Westbrook, University of New Hampshire <anthony.westbrook@unh.edu>
#
#   SPDX-License-Identifier: GPL-3.0-only WITH LicenseRef-repeatfs-graphviz-linking-source-exception
#


import os
import threading
from repeatfs.file_entry import FileEntry


class DescriptorEntry:
    """ Provides file descriptor information """
    _desc_lookup = dict()
    _file_lookup = dict()
    _pipe_lookup = dict()
    _idx = 1
    _lock = threading.RLock()

    @classmethod
    def get(cls, desc_id):
        """ Get a descriptor entry """
        with cls._lock:
            return cls._desc_lookup.get(desc_id, None)

    @classmethod
    def gen_pipe(cls, pipe, core):
        """ Make pipe, or retrieve pipe descriptor if exists already """
        with cls._lock:
            if pipe in cls._pipe_lookup:
                # Retrieve entry
                desc_entry = cls._desc_lookup[cls._pipe_lookup[pipe]]
            else:
                # Create and register new entry
                file_entry = FileEntry(pipe, core)
                desc_entry = DescriptorEntry(file_entry, None, core)
                cls._pipe_lookup[pipe] = desc_entry.id

        return desc_entry

    @classmethod
    def rename(cls, old_abs, new_virt, core):
        """ Update existing descriptors with new path information """
        with DescriptorEntry._lock:
            if old_abs not in cls._file_lookup:
                return

            file_entry = FileEntry(new_virt, core)
            cls._file_lookup.setdefault(file_entry.paths["abs_real"], set())

            # Update any open descriptors with new path info
            for desc_id in cls._file_lookup[old_abs]:
                cls.get(desc_id).file_entry = file_entry
                cls._file_lookup[file_entry.paths["abs_real"]].add(desc_id)

            del cls._file_lookup[old_abs]

    def __init__(self, file_entry, flags, core):
        self.core = core
        self.file_entry = file_entry
        self.flags = flags
        self.open_pid = self.core.fuse.fuse_get_context()[2]

        # For non-derived/api file, register real descriptor
        self.fs_lock = threading.RLock()
        self.fs_descriptor = None
        if not file_entry.derived_source and not file_entry.api and flags is not None:
            self.fs_descriptor = os.open(file_entry.paths["abs_real"], flags)

        # Generate thread-safe descriptor index and register
        with DescriptorEntry._lock:
            self.id = DescriptorEntry._idx
            DescriptorEntry._idx += 1
            DescriptorEntry._desc_lookup[self.id] = self
            DescriptorEntry._file_lookup.setdefault(file_entry.paths["abs_real"], set()).add(self.id)

    def __enter__(self):
        """ Allow with block for temporary descriptors """
        return self

    def __exit__(self, etype, evalue, trace):
        """ Remove temporary file descriptor """
        self.remove()

    def __str__(self):
        return "id {0} file_entry {1} flags {2} open_pid {3} fs_desc {4}".format(
            self.id, self.file_entry, self.flags, self.open_pid, self.fs_descriptor)

    def remove(self):
        """ Remove descriptor from descriptor and file lookups """
        with DescriptorEntry._lock:
            DescriptorEntry._file_lookup[self.file_entry.paths["abs_real"]].remove(self.id)
            if not DescriptorEntry._file_lookup[self.file_entry.paths["abs_real"]]:
                del DescriptorEntry._file_lookup[self.file_entry.paths["abs_real"]]

            del DescriptorEntry._desc_lookup[self.id]
